# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Device Power Default Capability."""
import time

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import device_power_base
from gazoo_device.utility import deprecation_utils

logger = gdm_logger.get_logger()

SUPPORTED_HUB_TYPES = [
    "cambrionix", "powerswitch", "unifi_switch", "ethernet_switch"
]


class DevicePowerDefault(device_power_base.DevicePowerBase):
  """Base class for device_power."""

  def __init__(self,
               device_name,
               create_device_func,
               hub_type,
               props,
               settable,
               hub_name_prop,
               port_prop,
               wait_for_bootup_complete_fn,
               switchboard_inst,
               change_triggers_reboot=False):
    """Create an instance of the device_power capability.

    Args:
        device_name (str): name of the device this capability is attached
          to.
        create_device_func (func): create_device method.
        hub_type (str): type of switch for power cycling.
        props (dict): dictionary of device props from configuration file.
        settable (bool): whether or not the properties are settable.
        hub_name_prop (str):  name of the hub name property
        port_prop (str): name of the hub port property
        wait_for_bootup_complete_fn (func): a method that the capability can
          call to wait for a reboot to complete if triggered by a change.
        switchboard_inst (object): instance of switchboard capability
        change_triggers_reboot (bool): set change_triggers_reboot to TRUE if
          changing the power mode for the device causes a reboot.

    Raises:
        ValueError: if hub type not a known type.
    """
    super().__init__(device_name=device_name)

    self._hub = None
    self._create_device_func = create_device_func
    self._settable = settable
    self._hub_name_prop = hub_name_prop
    self._props = props
    self._port_prop = port_prop
    if hub_type not in SUPPORTED_HUB_TYPES:
      raise ValueError("Hub Type {} not supported. Valid types: {}".format(
          hub_type, SUPPORTED_HUB_TYPES))
    self._hub_type = hub_type

    # Set the properties
    self._dict_name = "optional" if self._settable else "persistent_identifiers"

    self._wait_for_bootup_complete_fn = wait_for_bootup_complete_fn
    self._switchboard = switchboard_inst
    self._change_triggers_reboot = change_triggers_reboot

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self):
    """Checks that the capability is ready to use.

    Raises:
        CapabilityNotReadyError: if unable to create auxiliary device for
        power switching.
    """
    unset_props = []
    if self.hub_name is None:
      unset_props.append(self._hub_name_prop)
    if self.port_number is None:
      unset_props.append(self._port_prop)
    if unset_props:
      if self._settable:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm set-prop {} <property> <value>'")
      else:
        msg_format = ("If device is connected to {}, "
                      "set them via 'gdm redetect {}")
      msg = msg_format.format(self._hub_type, self._device_name)
      error_msg = "properties {} are unset. ".format(
          " and ".join(unset_props)) + msg
      raise errors.CapabilityNotReadyError(
          msg=error_msg, device_name=self._device_name)
    try:
      self._hub = self._create_device_func(self.hub_name)
    except errors.DeviceError as err:
      raise errors.CapabilityNotReadyError(
          msg=str(err), device_name=self._device_name)
    self._healthy = True

  @decorators.OptionalProperty
  def hub_name(self):
    """Name of the hub the device is attached to."""
    return self._props[self._dict_name].get(self._hub_name_prop)

  @decorators.PersistentProperty
  def hub_type(self):
    """Device type of the hub."""
    return self._hub_type

  @decorators.DynamicProperty
  def port_mode(self):
    """port mode."""
    if not self.healthy:
      self.health_check()
    return self._hub.switch_power.get_mode(self.port_number)

  @decorators.OptionalProperty
  def port_number(self):
    """Port number the device is attached to."""
    return self._props[self._dict_name].get(self._port_prop)

  @decorators.CapabilityLogDecorator(logger)
  def cycle(self, no_wait=False):
    """Power off then power on the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """
    if not self.healthy:
      self.health_check()
    self.off()
    time.sleep(2)  # Small delay before calling power_on
    self.on(no_wait=no_wait)

  @decorators.CapabilityLogDecorator(logger)
  def off(self, close_transports=True):
    """Turn off power to the device.

    Args:
      close_transports (bool): Whether to close transports when powering off.
    """
    if not self.healthy:
      self.health_check()
    if self._change_triggers_reboot:
      self._switchboard.add_log_note(
          f"GDM triggered reboot via {self.hub_type} power change.")
      self._hub.switch_power.power_off(self.port_number)
      self._wait_for_bootup_complete_fn()
    else:
      if close_transports:
        self._switchboard.close_all_transports()
      self._hub.switch_power.power_off(self.port_number)

  @decorators.CapabilityLogDecorator(logger)
  def on(self, no_wait=False):
    """Turn on power to the device.

    Args:
        no_wait (bool):  Return before verifying boot up.
    """
    if not self.healthy:
      self.health_check()
    if self._change_triggers_reboot:
      self._switchboard.add_log_note(
          f"GDM triggered reboot via {self.hub_type} power change.")
    self._hub.switch_power.power_on(self.port_number)
    if not self._change_triggers_reboot:
      time.sleep(2)  # Small delay to give time for 'dev/tty' to populate
      self._switchboard.open_all_transports()

    if not no_wait:
      self._wait_for_bootup_complete_fn()


deprecation_utils.add_deprecated_attributes(DevicePowerDefault,
                                            [("power_off", "off", True),
                                             ("power_on", "on", True),
                                             ("power_cycle", "cycle", True)])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transport based on the websocket protocol.

NOTE: the WebSocket protocol is NOT the same thing as a TCP/IP socket.
The WebSocket protocol is an application layer (layer 7) protocol.
    See https://tools.ietf.org/html/rfc6455.
TCP socket is a transport layer (layer 4) protocol.
    For a transport based on a TCP socket, see tcp_transport.py.
"""
import traceback

from gazoo_device import gdm_logger
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard.transports import tcp_transport
from gazoo_device.switchboard.transports import transport_base
import websocket

logger = gdm_logger.get_logger(__file__)


class WebSocketTransport(transport_base.TransportBase):
  """WebSocket-based transport."""

  def __init__(self,
               comms_address,
               connect_timeout=tcp_transport.CONNECT_TIMEOUT,
               auto_reopen=False,
               open_on_start=True):
    """Initialize the WebSocketTransport.

    Args:
        comms_address (str): websocket URL to connect to (ws://...)
        connect_timeout (float): timeout on the socket instance before
          attempting connect.
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.
    """
    super(WebSocketTransport, self).__init__(auto_reopen, open_on_start)
    self._properties.update({
        transport_properties.CONNECT_TIMEOUT: connect_timeout,
        transport_properties.WEBSOCKET_URL: comms_address
    })
    self.comms_address = comms_address
    self._websocket = None

  def is_open(self):
    """Returns True if websocket has been created and is not closed.

    Returns:
        bool: True if _websocket is not None
    """
    return hasattr(self, "_websocket") and self._websocket is not None

  def _open(self):
    """Opens or reopens the websocket connection using the current property values.

    Raises:
        Exception: if connection could not be established.
    """
    self._websocket = websocket.create_connection(
        self._properties[transport_properties.WEBSOCKET_URL],
        self._properties[transport_properties.CONNECT_TIMEOUT])

  def _close(self):
    """Closes the websocket connection."""
    self._websocket.close()
    self._websocket = None

  def _read(self, size=None, timeout=None):
    """Receives data from the websocket.

    Args:
        size (int): not used.
        timeout (float): maximum seconds to wait to read bytes or
          indefinitely if timeout=None. Default: None

    Returns:
        str: bytes read from transport
             or "" if no bytes were read
             or "" if an Exception occurred (including timeout)
    """
    self._websocket.settimeout(timeout)
    try:
      read_bytes = self._websocket.recv()
      return read_bytes
    except websocket.WebSocketTimeoutException:
      # Note: timeouts should not be logged. Otherwise, transport processes will flood
      # the logs with timeout exceptions (because they keep trying to read in infinite loop)
      return ""
    except Exception:
      logger.debug(
          "_read() from websocket with URL {} failed due to an error.".format(
              self._properties[transport_properties.WEBSOCKET_URL]))
      logger.debug(traceback.format_exc())
      return ""

  def _write(self, data, timeout=None):
    """Writes the data to the websocket.

    Args:
        data (str): bytes to be written out within timeout seconds
        timeout (float): maximum seconds to wait to write bytes or
          indefinitely if timeout=None. Default: None

    Returns:
        int: number of bytes written (0 if any Exception occurs, including
        timeout)
    """
    self._websocket.settimeout(timeout)
    try:
      bytes_sent = self._websocket.send(data)
      return bytes_sent
    except websocket.WebSocketTimeoutException:
      # Note: timeouts should not be logged. Otherwise, transport processes will flood
      # the logs with timeout exceptions (because they keep trying to read in infinite loop)
      return 0
    except Exception:
      logger.debug(
          "_write() to websocket with URL {} failed due to an error.".format(
              self._properties[transport_properties.WEBSOCKET_URL]))
      logger.debug(traceback.format_exc())
      return 0



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The implementation of the package management capability on Android devices.

The Android package management capability is used on the android devices which
can have
Android APKs installed. The Android package systems rely on adb command to
manage packages.
"""
from gazoo_device import decorators
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import package_management_base
from gazoo_device.utility import adb_utils

logger = gdm_logger.get_logger()

COMMAND_LIST_PACKAGES = 'pm list packages'


class PackageManagementAndroid(package_management_base.PackageManagementBase):
  """Android package management capability implementation."""

  def __init__(self, device_name, adb_serial):
    """Create a instance of the Android package management capability.

    Args:
        device_name (str): Name of device used for logging
        adb_serial (str): the device serial.
    """
    super(PackageManagementAndroid, self).__init__(device_name=device_name)
    self._adb_serial = adb_serial

  @decorators.CapabilityLogDecorator(logger)
  def install_package(self,
                      package_path,
                      allow_downgrade=False,
                      allow_test_apk=False,
                      reinstall=False):
    """Installs an Android package on the device.

    Args:
        package_path (str): the path to the package on the host machine.
        allow_downgrade (bool): allows version code downgrade.
        allow_test_apk (bool): allows test APKs to be installed.
        reinstall (bool): reinstalls an existing package and keeps its data.

    Raises:
        ValueError: when package_path is not valid.
        DeviceError: if failed to install the package.
    """
    adb_utils.install_package_on_device(
        package_path,
        adb_serial=self._adb_serial,
        allow_downgrade=allow_downgrade,
        allow_test_apk=allow_test_apk,
        reinstall=reinstall)

  @decorators.CapabilityLogDecorator(logger)
  def uninstall_package(self, package_name):
    """Uninstalls an Android package on the device.

    Args:
        package_name (str): the name of the package to uninstall.

    Raises:
        DeviceError: if failed to uninstall the package.
    """
    adb_utils.uninstall_package_on_device(
        package_name, adb_serial=self._adb_serial)

  @decorators.CapabilityLogDecorator(logger)
  def list_packages(self):
    """List all the installed packages on the device.

    Here we use `pm list package` command to show all the packages on the
    device. Note that the results will have be "package:<package_name>".
    We will remove the prefix here.

    Returns:
        list: package name strings.

    Raises:
        DeviceError: if failed to list the packages.
    """
    _PACKAGE_PREFIX = 'package:'

    cmd_result = adb_utils.shell(self._adb_serial, COMMAND_LIST_PACKAGES)
    packages = [
        pkg.replace(_PACKAGE_PREFIX, '') for pkg in cmd_result.split('\n')
    ]
    return packages



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the PwRPC (Pigweed RPC) lighting capability."""
from typing import Any, Callable, Tuple
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_light_base
from gazoo_device.switchboard.transports import pigweed_rpc_transport
try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  from lighting_service import lighting_service_pb2
  # pytype: enable=import-error
except ImportError:
  lighting_service_pb2 = None

logger = gdm_logger.get_logger()


class PwRPCLightDefault(pwrpc_light_base.PwRPCLightBase):
  """Pigweed RPC lighting capability for devices communicating over PwRPC."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any]):
    """Create an instance of the PwRPCLightDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The Switchboard.call method which calls to the endpoint.
        See more examples in nrf_pigweed_lighting.py.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call

  @decorators.CapabilityLogDecorator(logger)
  def on(self, no_wait: bool = False) -> None:
    """Turns on the light state of the device.

    Args:
      no_wait: If True, returns before verifying the light state.

    Raises:
      DeviceError: When the device does not transition to the appropriate
      state or if it remains off.
    """
    self._on_off(True, no_wait)

  @decorators.CapabilityLogDecorator(logger)
  def off(self, no_wait: bool = False) -> None:
    """Turns off the light state of the device.

    Args:
      no_wait: If True, returns before verifying the light state.

    Raises:
      DeviceError: When the device does not transition to the appropriate
      state or if it remains on.
    """
    self._on_off(False, no_wait)

  @decorators.DynamicProperty
  def state(self) -> bool:
    """The light state of the device.

    Returns:
      True if the device is in on state, false if it's in off state.

    Raises:
      DeviceError: When the device does not transition to the appropriate
      state or if it remains off.
    """
    state_in_bytes = self._get_state()
    state = lighting_service_pb2.LightingState.FromString(state_in_bytes)
    return state.on

  @decorators.DynamicProperty
  def brightness(self) -> Tuple[int, int]:
    """The brightness level of the device.

    Returns:
      Current brightness level and the maximal brightness level.
    """
    state_in_bytes = self._get_state()
    brightness = lighting_service_pb2.LightingBrightness.FromString(
        state_in_bytes)
    return brightness.level, brightness.max_level

  def _on_off(self, on: bool, no_wait: bool = False) -> None:
    """Turn on/off the light of the device.

    Args:
      on: Turns on the light if true, turn off the light otherwise.
      no_wait: If True, returns before verifying the light state.

    Raises:
      DeviceError: When the device does not transition to the appropriate
      state or if it remains the same state.
    """
    lighting_ack, _ = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=("Lighting", "Set"),
        method_kwargs={"on": on})
    action = "on" if on else "off"
    if not lighting_ack:
      raise errors.DeviceError(
          f"Device {self._device_name} turning light {action} failed.")
    if not no_wait:
      if on != self.state:  # pylint: disable=comparison-with-callable
        raise errors.DeviceError(
            f"Device {self._device_name} light didn't turn {action}.")

  def _get_state(self) -> bytes:
    """Returns the serialized lighting state of the device.

    Returns:
      The serialized lighting state.

    Raises:
      DeviceError: When the ack value is not true.
    """
    ack, state_in_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=("Lighting", "Get"),
        method_kwargs={})
    if not ack:
      raise errors.DeviceError(
          f"Device {self._device_name} getting light state failed.")
    return state_in_bytes



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Communication Power Cambrionix Capability.

Note: this should be refactored to be multi communication ports capability
if/when another device starts needing it.
"""
import time
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import comm_power_base
from gazoo_device.capabilities.interfaces import switch_power_base
from gazoo_device.utility import deprecation_utils

logger = gdm_logger.get_logger()


class CommPowerUsbEthernet(comm_power_base.CommPowerBase):
  """Base class for comm_power  capability."""

  def __init__(self,
               device_name,
               create_device_func,
               props,
               hub_name_prop,
               primary_port_prop,
               secondary_port_prop,
               ethernet_switch_prop,
               ethernet_port_prop,
               get_switchboard_if_initialized,
               power_and_data_share_cable=False,
               pre_off_func=None):
    """Create an instance of the comm_power_cambrionix capability.

    Args:
      device_name (str): name of the device this capability is attached
        to.
      create_device_func (func): create_device method.
      props (dict): dictionary of device props from configuration file.
      hub_name_prop (str):  name of the hub name property. Should be in
        optional.
      primary_port_prop (str): name of the hub port property. Should be in
        optional.
      secondary_port_prop (str): name of the secondary port property in
        persistent.
      ethernet_switch_prop (str): ip address of the ethernet_switch prop.
        Optional.
      ethernet_port_prop (str): port number of the ethernet_switch port.
        Optional.
      get_switchboard_if_initialized (callable): function which returns
        a Switchboard instance or None if Switchboard hasn't been initialized.
      power_and_data_share_cable (bool): set power_and_data_share_cable to
        TRUE if device shares a common USB cable for its power and data.
      pre_off_func (func): a method to call before turning power off.

    Raises:
      ValueError: If the capability cannot be created.
    """
    super().__init__(device_name=device_name)

    self._create_device_func = create_device_func
    self._hub_name_prop = hub_name_prop
    self._primary_port_prop = primary_port_prop
    self._secondary_port_prop = secondary_port_prop
    self._props = props
    self._ethernet_switch = None

    # Set the properties
    self._get_switchboard_if_initialized = get_switchboard_if_initialized
    self._power_and_data_share_cable = power_and_data_share_cable
    self._pre_off_func = pre_off_func
    self._ethernet_switch_prop = ethernet_switch_prop
    self._ethernet_port_prop = ethernet_port_prop

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def health_check(self):
    """Checks that the capability is ready to use.

    Raises:
        CapabilityNotReadyError: if unable to create auxiliary device for
        power switching.
    """
    unset_props = []
    if not self.hub_name:
      unset_props.append(self._hub_name_prop)
    if not self.port_number:
      unset_props.append(self._primary_port_prop)
    if unset_props:
      msg_format = ("If device is connected to {}, "
                    "set them via 'gdm redetect {}'")
      msg = msg_format.format(self.hub_type, self._device_name)
      error_msg = "properties {} are unset. ".format(
          " and ".join(unset_props)) + msg
      raise errors.CapabilityNotReadyError(
          msg=error_msg, device_name=self._device_name)

    try:
      self._hub = self._create_device_func(self.hub_name)
      # Set up ethernet
      if self.ethernet_switch_address is not None:
        self._ethernet_switch = self._create_device_func(
            self.ethernet_switch_address)

    except errors.DeviceError as err:
      raise errors.CapabilityNotReadyError(
          msg=str(err), device_name=self._device_name)
    if self.ethernet_switch_address is not None:
      self._verify_switch_created(self._ethernet_switch)
    self._healthy = True

  @decorators.OptionalProperty
  def hub_name(self):
    """Name of the hub the device is attached to."""
    return self._props["persistent_identifiers"].get(self._hub_name_prop)

  @decorators.DynamicProperty
  def port_mode(self):
    """port mode."""
    if not self.healthy:
      self.health_check()
    return self._hub.switch_power.get_mode(self.port_number)

  @decorators.OptionalProperty
  def port_number(self):
    """Port number the device is attached to."""
    return self._props["persistent_identifiers"].get(self._primary_port_prop)

  @decorators.OptionalProperty
  def secondary_port_number(self):
    """Secondary usb port number that the device is attached to."""
    return self._props["persistent_identifiers"].get(self._secondary_port_prop)

  @decorators.OptionalProperty
  def ethernet_switch_address(self):
    """Address for ethernet switch."""
    return self._props["optional"].get(self._ethernet_switch_prop)

  @decorators.OptionalProperty
  def ethernet_port_number(self):
    """Port number for ethernet switch."""
    return self._props["optional"].get(self._ethernet_port_prop)

  @decorators.PersistentProperty
  def hub_type(self):
    """Type of hub the check is attached to."""
    return "cambrionix"

  @decorators.CapabilityLogDecorator(logger)
  def cycle(self, include_ethernet=False):
    """Power off then power on the communication port."""
    if not self.healthy:
      self.health_check()
    self.off()
    time.sleep(2)  # Small delay before calling power_on
    self.on()

  @decorators.CapabilityLogDecorator(logger)
  def off(self, include_ethernet=False):
    """Turn off power to the device communications port."""
    if not self.healthy:
      self.health_check()
    if self._pre_off_func:
      self._pre_off_func()
    switchboard = self._get_switchboard_if_initialized()
    if self._power_and_data_share_cable:
      if switchboard:
        switchboard.add_log_note(
            f"comm_power.off() called on {self._device_name} set communication "
            f"port {self.port_number} to charge as device has a single USB "
            "cable for data and power.")
        switchboard.close_all_transports()
      self._hub.switch_power.power_on(self.port_number, data_sync=False)
      if self.secondary_port_number is not None:
        self._hub.switch_power.power_on(
            self.secondary_port_number, data_sync=False)
    else:
      if switchboard:
        switchboard.close_all_transports()
      self._hub.switch_power.power_off(self.port_number)
      if self.secondary_port_number is not None:
        self._hub.switch_power.power_off(self.secondary_port_number)
    if include_ethernet:
      self.ethernet_off()

  @decorators.CapabilityLogDecorator(logger)
  def on(self, include_ethernet=False):
    """Turn on power to the device communication port."""
    if not self.healthy:
      self.health_check()
    self._hub.switch_power.power_on(self.port_number)
    if self.secondary_port_number is not None:
      self._hub.switch_power.power_on(self.secondary_port_number)
    if include_ethernet:
      self.ethernet_on()
    time.sleep(5)  # Small delay to give time for 'dev/tty' to populate
    switchboard = self._get_switchboard_if_initialized()
    if switchboard:
      switchboard.open_all_transports()

  @decorators.CapabilityLogDecorator(logger)
  def ethernet_off(self):
    """Turn off ethernet port.

    Raises:
      CapabilityNotReadyError: if unable to initiate ethernet_switch.
    """
    if not self.healthy:
      self.health_check()
    if not self._ethernet_switch:
      raise errors.CapabilityNotReadyError(
          device_name=self._device_name,
          msg="Not set up for ethernet switching.")
    self._ethernet_switch.switch_power.power_off(self.ethernet_port_number)

  @decorators.CapabilityLogDecorator(logger)
  def ethernet_on(self):
    """Turn on ethernet port.

    Raises:
      CapabilityNotReadyError: if unable to initiate ethernet_switch.
    """
    if not self.healthy:
      self.health_check()
    if not self._ethernet_switch:
      raise errors.CapabilityNotReadyError(
          device_name=self._device_name,
          msg="Not set up for ethernet switching.")
    self._ethernet_switch.switch_power.power_on(self.ethernet_port_number)

  def _verify_switch_created(self, switch):
    """Verifies switch is created and has switch_power capability."""
    if not (
        hasattr(switch, "switch_power") and
        isinstance(switch.switch_power, switch_power_base.SwitchPowerBase)):
      raise errors.CapabilityNotReadyError(
          msg="'switch_power' capability is missing in hub device {} ({}),"
          " or is not an instance of SwitchPowerBase".format(
              self.hub_name,
              type(switch).__name__),
          device_name=self._device_name)


deprecation_utils.add_deprecated_attributes(
    CommPowerUsbEthernet, [("enable", "ethernet_on", True),
                           ("disable", "ethernet_off", True),
                           ("address", "ethernet_switch_address", False),
                           ("port", "ethernet_port_number", False),
                           ("power_cycle", "cycle", True)])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Switchboard capability implementation.

Switchboard is the backbone of all device interaction.
It provides the ability to interact with devices using standardized transport,
button, and expect APIs.

By separating these standardized APIs we can more easily test the logic and
eventually unit test device classes independent of hardware.
"""
import io
import multiprocessing
import os
import queue
import re
import signal
import subprocess
import time
import types
from typing import Any, Dict, List, Optional, Tuple

import xmodem

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import switchboard_base
from gazoo_device.switchboard import line_identifier
from gazoo_device.switchboard import log_process
from gazoo_device.switchboard import switchboard_process
from gazoo_device.switchboard import transport_process
from gazoo_device.switchboard import transport_properties
from gazoo_device.switchboard.expect_response import ExpectResponse
from gazoo_device.switchboard.transports import jlink_transport
from gazoo_device.switchboard.transports import serial_transport
from gazoo_device.switchboard.transports import tcp_transport
from gazoo_device.switchboard.transports import transport_base
from gazoo_device.utility import common_utils
from gazoo_device.utility import usb_utils

logger = gdm_logger.get_logger("core")

MODE_TYPE_ALL = switchboard_base.MODE_TYPE_ALL
MODE_TYPE_ANY = switchboard_base.MODE_TYPE_ANY
MODE_TYPE_SEQUENTIAL = switchboard_base.MODE_TYPE_SEQUENTIAL
VERIFY_METHOD_MD5SUM = switchboard_base.VERIFY_METHOD_MD5SUM
_VALID_EXPECT_TYPES = [
    line_identifier.LINE_TYPE_ALL, line_identifier.LINE_TYPE_LOG,
    line_identifier.LINE_TYPE_RESPONSE
]
_VALID_EXPECT_MODES = [MODE_TYPE_ALL, MODE_TYPE_ANY, MODE_TYPE_SEQUENTIAL]
_VERIFY_METHODS = [VERIFY_METHOD_MD5SUM]


def _ensure_has_newline(cmd, add_newline=True, newline="\n"):
  """Ensure command ends with a newline (if add_newline is True)."""
  if add_newline and not cmd.endswith(("\n", "\x00", r"\0", newline)):
    cmd += newline
  return cmd


def _get_pattern_index(compiled_list, match_list, mode):
  """Return index of compiled regex pattern that matches the match provided.

  Args:
      compiled_list (list): of regular expression patterns to match
      match_list (list): of Match objects previously found
      mode (str): type of expect to use ("any", "all", or "sequential")

  Returns:
      int: The index to a matching compile regex pattern in compiled_list
           that matches the last entry added to match_list or None if no
           match was found.
  """

  if match_list:
    if mode == MODE_TYPE_SEQUENTIAL:
      return len(match_list) - 1
    else:
      for index, pattern in enumerate(compiled_list):
        if pattern == match_list[-1].re:
          return index
  return None


def _get_pattern_list(compiled_list, match_list, mode):
  """Returns pattern_list to be used in expect search using information provided.

  Args:
      compiled_list (list): of regular expression patterns
      match_list (list): of Match objects previously found
      mode (str): type of expect to use ("any", "all", or "sequential")

  Returns:
      list: A list of compiled regular expression patterns from compiled_list
            to use for the next round of pattern searches according to the
            mode specified.

  Note:
      The primary difference between expect type modes is what pattern list
      should be used. For any, the whole list should be returned every time
      until one match is found. For all, only the remaining unmatched
      patterns should be returned. For sequential, only the first of the
      remaining unmatched patterns should be returned.
  """

  if mode == MODE_TYPE_ANY:
    return compiled_list
  missing_patterns = _get_missing_patterns(compiled_list, match_list, mode)
  if mode == MODE_TYPE_ALL:
    return missing_patterns
  return [missing_patterns[0]]


def _get_missing_patterns(compiled_list, match_list, mode):
  """Returns compiled regex patterns from compiled_list that are not in match_list.

  Args:
      compiled_list (list): of regular expression patterns
      match_list (list): of Match objects previously found
      mode (str): type of expect to use ("any", "all", or "sequential")

  Returns:
      list: A list of compiled regex patterns from compiled_list that are not
            in match_list.
  """
  if mode == MODE_TYPE_SEQUENTIAL:
    return compiled_list[len(match_list):]
  else:
    matched_patterns = [match.re for match in match_list]
    return [
        pattern for pattern in compiled_list if pattern not in matched_patterns
    ]


def _get_pattern_strings(compiled_list):
  """Returns regex pattern strings from a list of compiled regex pattern objects.

  Args:
      compiled_list (list): of regular expression patterns to extract strings
        from

  Returns:
      list: A list of regex pattern strings extracted from compiled list of
            regex pattern objects.
  """
  return [pattern.pattern for pattern in compiled_list]


class SwitchboardDefault(switchboard_base.SwitchboardBase):
  """Manages device interactions and writes everything to a single file.

  This will spawn and manage 3 or more multiprocess subprocesses:
      Process 0 (Main)
          The primary process where the API/CLI/tests execute from;
          responsible for initializing this module for each device. Performs
          expect and Parser logic and closes subprocesses on shutdown or
          transport errors.

      Process 1..n (Transport)
          One or more subprocesses responsible for performing all device
          transport operations (e.g. open/close and read/write) and
          communicating results to other subprocesses as needed using
          queues.

      Process n+1 (Log writer)
          Performs all log writing operations received from log queue
          shared with Transport subprocess. Log lines are written only for
          completed log lines.

      Process n+2 (Log filter)
          Reads log lines from the log file written by the log writer
          subprocess. Filters each log line read for desired events and
          writes them to an event file. The main process can then use the
          event file to query for relevant events.
  """

  def __init__(
      self,
      device_name,
      exception_queue,
      transport_list,
      log_path,
      framer_list=None,
      identifier=None,
      button_list=None,
      parser=None,
      partial_line_timeout_list=None,
      force_slow=False,
      max_log_size=0,
  ):
    """Initialize the Switchboard with the parameters provided.

    Args:
      device_name (str): name of the device
      exception_queue (Queue): a single exception queue to share for all
        subprocesses to report exceptions into
      transport_list (list): of Transport type classes to use for this
        device.
      log_path (str): path/to/log/file to store device logs
      framer_list (list): list of DataFramer derived classes to use to
        frame incoming raw data into raw lines.
      identifier (LineIdentifier): LineIdentifier derived class to use to
        identify log and response lines from device for expect_type
        filtering.
      button_list (list): of Button type classes to use to click, press,
        release, is_valid, valid_buttons, and close methods defined for
        button instigation, button name verification and listing, and
        closing.
      parser (Parser): log filtering object which must have
        "process_line".
      partial_line_timeout_list (list): of seconds to wait before writing
        partial transport line received.
      force_slow (bool): flag indicating all sends should assume
        slow=True.
      max_log_size (int): maximum size in bytes before performing log
        rotation. max_log_size of 0 means no log rotation should ever occur.
    """
    super().__init__(device_name=device_name)
    if framer_list is None:
      framer_list = []
    if partial_line_timeout_list is None:
      partial_line_timeout_list = []

    self.log_path = log_path
    self._button_list = button_list
    self._force_slow = force_slow
    self._identifier = identifier or line_identifier.AllUnknownIdentifier()
    time.sleep(.1)
    common_utils.run_before_fork()
    self._mp_manager = multiprocessing.Manager()
    common_utils.run_after_fork_in_parent()
    self._transport_processes = []
    self._log_queue = self._mp_manager.Queue()
    self._call_result_queue = self._mp_manager.Queue()
    self._raw_data_queue = self._mp_manager.Queue()
    self._raw_data_queue_users = 0
    self._transport_process_id = 0
    self._exception_queue = exception_queue

    self._add_transport_processes(transport_list, framer_list,
                                  partial_line_timeout_list)
    self._add_log_writer_process(log_path, max_log_size)
    self._add_log_filter_process(parser, log_path)
    self._start_processes()

  def __del__(self):
    self.close()

  @decorators.CapabilityLogDecorator(logger, level=None)
  def add_log_note(self, note):
    """Adds given note to device log file.

    Args:
        note (str): to write to the log file
    """

    if note[-1] == "\n":
      log_message = "Note: {}".format(note)
    else:
      log_message = "Note: {}\n".format(note)
    try:
      log_process.log_message(self._log_queue, log_message, "M")
    except (AttributeError, IOError):  # manager shutdown or close called
      pass

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def add_new_filter(self, filter_path):
    """Adds new log filter at path specified to LogFilterProcess.

    Args:
        filter_path (str): filter file to add

    Raises:
        RuntimeError: if LogFilterProcess is not available or running.
        ValueError: if filter_path doesn't exist
    """
    if not os.path.exists(filter_path):
      raise ValueError("Filter path {} doesn't exist.".format(filter_path))

    if not hasattr(
        self,
        "_log_filter_process") or not self._log_filter_process.is_running():
      raise RuntimeError("Log filter process is not currently running.")

    self._log_filter_process.send_command(log_process.CMD_ADD_NEW_FILTER,
                                          filter_path)
    while (self._log_filter_process.is_running() and
           not self._log_filter_process.is_command_done()):
      time.sleep(0.001)

  def call(self,
           method: types.MethodType,
           method_args: Tuple[Any, ...] = (),
           method_kwargs: Optional[Dict[str, Any]] = None,
           port: int = 0) -> Any:
    """Calls a transport method in a transport process and returns the response.

    Args:
      method: the transport method to execute.
      method_args: positional arguments for the call.
      method_kwargs: keyword arguments for the call.
      port: number of the transport to call the method in.

    Raises:
      DeviceError: mismatching transport type.
      Exception: exceptions encountered in the transport process are reraised.

    Returns:
      object: return value of the transport method.

    Note that the call is executed in a different process. Therefore all
    transport method arguments and the return value must be serializable.
    """
    method_kwargs = method_kwargs or {}
    self._validate_port(port, self.call.__name__)
    class_name = method.__qualname__.split(".")[-2]
    method_name = method.__qualname__.split(".")[-1]
    transport_class_name = type(
        self._transport_processes[port].transport).__name__
    if (class_name != transport_base.TransportBase.__name__
        # All transports inherit from TransportBase
        and class_name != transport_class_name):
      raise errors.DeviceError(
          f"{self._device_name} Switchboard.call failed. "
          f"Requested method {method.__qualname__!r}, but transport {port} "
          f"is of type {transport_class_name!r}.")
    self.add_log_note("Executing {!r} in transport {}"
                      .format(method.__qualname__, port))
    self._transport_processes[port].send_command(
        transport_process.CMD_TRANSPORT_CALL,
        (method_name, method_args, method_kwargs))
    success, response = self._call_result_queue.get()
    if success:
      return response
    raise errors.DeviceError(
        f"{self._device_name} switchboard.call of method {method.__qualname__} "
        f"in transport {port} failed. {response}")

  def call_and_expect(self,
                      method: types.MethodType,
                      pattern_list: List[str],
                      timeout: float = 30.0,
                      searchwindowsize: int = config.SEARCHWINDOWSIZE,
                      expect_type: str = line_identifier.LINE_TYPE_ALL,
                      mode: str = MODE_TYPE_ANY,
                      method_args: Tuple[Any, ...] = (),
                      method_kwargs: Optional[Dict[str, Any]] = None,
                      port: int = 0,
                      raise_for_timeout: bool = False):
    """Calls a transport method and expects on the patterns provided.

    Args:
        method: The transport method to execute.
        pattern_list: List of regex expressions to look for in the lines.
        timeout: Seconds to look for the patterns.
        searchwindowsize: Number of the last bytes to look at.
        expect_type: 'log', 'response', or 'all'.
        mode: Type of expect to run ("any", "all" or "sequential").
        method_args: Positional arguments for the call.
        method_kwargs: Keyword arguments for the call.
        port: Number of the transport to call the method in.
        raise_for_timeout: Raise an exception if the expect times out.

    Raises:
        DeviceError: if port specified or other expect arguments are
                     invalid, or timed out.

    Returns:
        tuple: (ExpectResponse, object)
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects
        object: Returned value of the transport method.
    """
    expect_ret, func_ret = self.do_and_expect(
        self.call, [method], {
            "method_args": method_args,
            "method_kwargs": method_kwargs,
            "port": port
            },
        pattern_list,
        timeout=timeout,
        searchwindowsize=searchwindowsize,
        expect_type=expect_type,
        mode=mode,
        include_func_response=True)
    if expect_ret and expect_ret.timedout and raise_for_timeout:
      raise errors.DeviceError(
          "Device {} call_and_expect timed out for method {} in {}s".format(
              self._device_name, method.__name__, timeout))
    return expect_ret, func_ret

  @decorators.CapabilityLogDecorator(logger)
  def click(self, button, duration=0.5, port=0):
    """Press and release the button for the duration and port specified.

    Args:
        button (str): button to press and release
        duration (float): seconds to wait before releasing button
        port (int): which port to click on, 0 or 1.

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, duration, or port values are invalid
    """
    self._check_button_args("click", button, port, duration=duration)

    log_message = "click button {} on port {} for duration {} - begin".format(
        button, port, duration)
    self.add_log_note(log_message)
    self._button_list[port].click(button, duration)
    log_message = "click button {} on port {} for duration {} - end".format(
        button, port, duration)
    self.add_log_note(log_message)

  def click_and_expect(self,
                       button,
                       pattern_list,
                       duration=0.5,
                       timeout=30.0,
                       searchwindowsize=config.SEARCHWINDOWSIZE,
                       expect_type="log",
                       port=0,
                       mode="any",
                       raise_for_timeout=False):
    """Press and release button, log lines matching patterns are returned.

    Args:
        button (str): button to press and release
        pattern_list (list): list of regex expressions to look for in the
          lines
        duration (int): seconds to press button before releasing it
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        port (int): which port to send on, 0 or 1
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out

    Raises:
        DeviceError: If buttons are not supported on the device or
                     other arguments are invalid.

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect. Starts up
        expect queue right before clicking button to catch fast responses.
    """
    return self.do_and_expect(
        self.click, [button], {
            "duration": duration,
            "port": port
        },
        pattern_list,
        timeout=timeout,
        searchwindowsize=searchwindowsize,
        expect_type=expect_type,
        mode=mode,
        raise_for_timeout=raise_for_timeout)

  @decorators.CapabilityLogDecorator(logger, level=None)
  def close(self):
    """Shuts down the subprocesses and closes the transports.

    NOTE:
        The current implementation relies on queues being garbage collected.
        Instead of explicitly closing the queues, all queue references MUST
        be deleted to
        release the queues and prevent a memory leak!
    """
    comms_addresses = [
        proc.transport.comms_address for proc in self._transport_processes
    ]
    self._stop_processes()
    if hasattr(self, "_button_list") and self._button_list:
      for button in self._button_list:
        button.close()
      self._button_list = []
    if hasattr(self, "_call_result_queue") and self._call_result_queue:
      delattr(self, "_call_result_queue")
    if hasattr(self, "_raw_data_queue") and self._raw_data_queue:
      delattr(self, "_raw_data_queue")
    if hasattr(self, "_log_queue") and self._log_queue:
      delattr(self, "_log_queue")
    if hasattr(self, "_exception_queue") and self._exception_queue:
      delattr(self, "_exception_queue")
    if hasattr(self, "_mp_manager") and self._mp_manager:
      self._mp_manager.shutdown()
      delattr(self, "_mp_manager")
    self.ensure_serial_paths_unlocked(comms_addresses)

  @decorators.CapabilityLogDecorator(logger)
  def close_all_transports(self):
    """Leaves the switchboard architecture intact but closes the communication FDs.

    This is used prior to the connections being closed,
    such as disconnecting an ethernet or a serial connection.
    Only closes the ones open so if device.close has already occurred,
    nothing will be closed.
    """
    for port in range(self.number_transports):
      self.close_transport(port=port)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def close_transport(self, port=0):
    """Closes the transport specified.

    Args:
        port (int or str): the transport port to close

    Raises:
        DeviceError: If port value is invalid or out of range.
    """
    self._validate_port(port, self.close_transport.__name__)

    log_message = "closing transport for port {}".format(port)
    self.add_log_note(log_message)
    try:
      if self._button_list:
        button = self._button_list[port]
        button.close()
    except IndexError:
      pass  # If _button_list does not have an entry for [port], there is nothing to close

    start_time = time.time()
    transport_proc = self._transport_processes[port]
    transport_proc.send_command(transport_process.CMD_TRANSPORT_CLOSE)
    while transport_proc.is_open():
      time.sleep(0.01)
    log_message = "closed transport for port {} in {}s".format(
        port,
        time.time() - start_time)
    self.add_log_note(log_message)

  def do_and_expect(self,
                    func,
                    func_args,
                    func_kwargs,
                    pattern_list,
                    timeout=30.0,
                    searchwindowsize=config.SEARCHWINDOWSIZE,
                    expect_type=line_identifier.LINE_TYPE_LOG,
                    mode=MODE_TYPE_ANY,
                    raise_for_timeout=False,
                    include_func_response=False):
    """Executes function with given args, blocks until expect matches or timeout occurs.

    Args:
        func (method): name of function to be called
        func_args (list): positional arguments specified to be passed to
          function
        func_kwargs (dict): keyword arguments specified to be passed to
          function
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out
        include_func_response (bool): If True, also return func's return value.

    Returns:
        ExpectResponse or tuple: ExpectResponse or (ExpectResponse, object) if
        include_func_response is True.

        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects
        object: Returned value of the transport method.

    Raises:
        DeviceError: If func is not callable
                     If other arguments are invalid

    Note:
        Input parameter "func" MUST NOT call "shell" nor another
        "core.xxx_expect" method so as to avoid the nested "flush"
        problem described in 'NEP-2343'.
    """
    if not callable(func):
      raise errors.DeviceError("Device {} do_and_expect failed. "
                               "Function: {} is not callable.".format(
                                   self._device_name, func))
    self._check_expect_args(pattern_list, timeout, searchwindowsize,
                            expect_type, mode)
    compiled_list = self._get_compiled_pattern_list(pattern_list)

    try:
      self._enable_raw_data_queue()
      func_ret = func(*func_args, **func_kwargs)
      expect_ret = self._expect(
          compiled_list,
          timeout,
          searchwindowsize,
          expect_type,
          mode,
          raise_for_timeout=raise_for_timeout)
      if include_func_response:
        return expect_ret, func_ret
      else:
        return expect_ret
    finally:
      self._disable_raw_data_queue()

  @decorators.CapabilityLogDecorator(logger)
  def echo_file_to_transport(self,
                             source_file,
                             destination_path,
                             port=0,
                             bytes_per_echo=50):
    r"""Transfers file to transport specified using echo commands.

    Args:
        source_file (path): to the file to transfer
        destination_path (path): to transfer file to on device
        port (int or str): the transport port to open
        bytes_per_echo (int): call to use during file transfer

    Raises:
        DeviceError: If source_file doesn't exist, can't be opened, or
                     the port or bytes_per_echo values are invalid or
                     out of range.

    Note:
        The caller is responsible for preparing the device to receive
        multiple echo commands to receive the file and only calling this
        method for devices that support the following commands::

            echo -ne > <destination_path>
            echo -ne "\\x{:02x}" >> <destination_path>
    """
    self._validate_port(port, self.echo_file_to_transport.__name__)

    if not os.path.exists(source_file):
      raise errors.DeviceError("Device {} echo file to transport failed. "
                               "Source file {} doesn't exist.".format(
                                   self._device_name, source_file))
    elif not isinstance(bytes_per_echo, int):
      raise errors.DeviceError(
          "Device {} echo file to transport failed. "
          "Expecting int for bytes_per_echo found {} instead.".format(
              self._device_name, type(bytes_per_echo)))
    elif bytes_per_echo <= 0:
      raise errors.DeviceError(
          "Device {} echo file to transport failed. "
          "Invalid bytes_per_echo value {} expected >0".format(
              self._device_name, bytes_per_echo))

    start_time = time.time()
    log_message = ("starting echo transfer of {} for port {} to {}".format(
        source_file, port, destination_path))
    self.add_log_note(log_message)

    # Read the data in from the file
    try:
      with io.open(source_file, "rb") as in_file:
        data = in_file.read()
    except IOError as err:
      raise errors.DeviceError("Device {} echo file to transport failed. "
                               "Unable to read {}. "
                               "Error: {!r}".format(self._device_name,
                                                    source_file, err))

    # Create and zero out the file
    cmd_string = "echo -ne > {}\n".format(destination_path)
    echo_pattern1 = r"echo -ne > [/\w.]+"
    result = self.send_and_expect(
        cmd_string, [echo_pattern1], port=port, expect_type="response")
    if result.timedout:
      raise errors.DeviceError("Device {} echo file to transport failed. "
                               "Unable to create/erase file {} on device. "
                               "Output: {!r}".format(self._device_name,
                                                     destination_path,
                                                     result.before))

    # Loop through the data at bytes_per_echo chunks at a time
    echo_pattern2 = r'echo -ne ".*"\s.*\s[/\w.]+'
    data_indexes = list(range(0, len(data), bytes_per_echo)) + [len(data)]
    for data_index in range(1, len(data_indexes)):
      start_index = data_indexes[data_index - 1]
      end_index = data_indexes[data_index]
      data_chunk = u"".join(
          [u"\\x%02x" % byte for byte in data[start_index:end_index]])
      cmd_string = "echo -ne \"{}\" >> {}\n".format(data_chunk,
                                                    destination_path)
      result = self.send_and_expect(
          cmd_string, [echo_pattern2], port=port, expect_type="response")
      if result.timedout:
        raise errors.DeviceError("Device {} echo file to transport failed. "
                                 "Unable to echo bytes {!r} to file {} "
                                 "Output: {!r}".format(self._device_name,
                                                       data_chunk,
                                                       destination_path,
                                                       result.before))
    log_message = ("finished echo transfer of {} for port {} in {}s".format(
        source_file, port,
        time.time() - start_time))
    self.add_log_note(log_message)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def ensure_serial_paths_unlocked(self, communication_addresses: List[str]):
    """Ensures serial paths are longer locked by switchboard process after device is closed."""
    for comms_address in communication_addresses:
      if comms_address not in usb_utils.get_all_serial_connections():
        continue  # only active serial connections have locks

      try:
        output = subprocess.check_output(["lsof", comms_address],
                                         stderr=subprocess.STDOUT)
        output = output.decode("utf-8", "replace").splitlines()
      except (subprocess.CalledProcessError,
              ImportError):  # process does not exist
        continue

      process_number = output[-1].split()[1]  # process id
      try:
        number = int(process_number)
        logger.info(
            "{}'s communication process still exists. Killing process {}"
            .format(self._device_name, number))
        os.kill(int(process_number), signal.SIGTERM)
      except ValueError:
        logger.error(
            "{}'s communication process still exists "
            "but could not parse process number from lsof output correctly. "
            "Output: {}. Expected process number: {}".format(
                self._device_name, output, process_number))

  def expect(self,
             pattern_list,
             timeout=30.0,
             searchwindowsize=config.SEARCHWINDOWSIZE,
             expect_type=line_identifier.LINE_TYPE_ALL,
             mode=MODE_TYPE_ANY,
             raise_for_timeout=False):
    """Block until a regex pattern is matched or until a timeout time has elapsed.

    Args:
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out

    Raises:
        DeviceError: if arguments are not valid.

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect.
    """
    self._check_expect_args(pattern_list, timeout, searchwindowsize,
                            expect_type, mode)
    compiled_list = self._get_compiled_pattern_list(pattern_list)

    try:
      self._enable_raw_data_queue()
      return self._expect(
          compiled_list,
          timeout,
          searchwindowsize,
          expect_type,
          mode,
          raise_for_timeout=raise_for_timeout)
    finally:
      self._disable_raw_data_queue()

  def get_line_identifier(self):
    """Returns the line identifier currently used by Switchboard."""
    return self._identifier

  @decorators.DynamicProperty
  def number_transports(self) -> int:
    """Returns the number of transport processes used by Switchboard."""
    return len(self._transport_processes)

  @decorators.CapabilityLogDecorator(logger)
  def open_all_transports(self):
    """Opens the communication FDs, assuming switchboard architecture is intact.

    This is used after a physical connection has been reopened,
    such as reconnecting an ethernet or a serial connection.
    Only opens the ones closed so if device.close has already occurred,
    nothing will be opened.
    """
    for port in range(self.number_transports):
      self.open_transport(port=port)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def open_transport(self, port=0, timeout=30.0):
    """Opens the transport specified.

    Args:
        port (int or str): the transport port to open
        timeout (float): how long to wait for port to open.

    Raises:
        DeviceError: If port value is invalid or out of range.
    """
    self._validate_port(port, self.open_transport.__name__)

    log_message = "opening transport for port {}".format(port)
    self.add_log_note(log_message)
    transport_proc = self._transport_processes[port]
    start_time = time.time()
    transport_proc.send_command(transport_process.CMD_TRANSPORT_OPEN)
    start_time = time.time()
    elapsed_time = 0
    while not transport_proc.is_open():
      if elapsed_time < timeout:
        time.sleep(0.01)
        elapsed_time = time.time() - start_time
      else:
        return
    log_message = "opened transport for port {} in {}s".format(
        port,
        time.time() - start_time)
    self.add_log_note(log_message)

  @decorators.CapabilityLogDecorator(logger)
  def press(self, button, wait=0.0, port=0):
    """Presses the button for the port specified and waits the time specified.

    Args:
        button (str): button to press
        wait (float): seconds to wait before returning
        port (int): which port to click on, 0 or 1

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, wait, or port values are invalid
    """
    self._check_button_args("press", button, port, wait=wait)

    log_message = "pressing button {} on port {} and waiting {}s - begin".format(
        button, port, wait)
    self.add_log_note(log_message)
    self._button_list[port].press(button, wait)
    log_message = "pressing button {} on port {} and waiting {}s - end".format(
        button, port, wait)
    self.add_log_note(log_message)

  def press_and_expect(self,
                       button,
                       pattern_list,
                       wait=0.0,
                       timeout=30.0,
                       searchwindowsize=config.SEARCHWINDOWSIZE,
                       expect_type="log",
                       port=0,
                       mode="any"):
    """Press button and expect for pattern_list and other arguments provided.

    Args:
        button (str): button to press
        pattern_list (list): list of regex expressions to look for in the
          lines
        wait (float): seconds to wait
        timeout (float): Seconds to look for the patterns
        searchwindowsize (int): Number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        port (int): Which port to send on, 0 or 1
        mode (str): type of expect to run ("any", "all" or "sequential")

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, wait, port, or expect values are invalid

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect. Starts up
        expect queue right before pressing button to catch fast responses.
    """
    return self.do_and_expect(
        self.press, [button], {
            "wait": wait,
            "port": port
        },
        pattern_list,
        timeout=timeout,
        searchwindowsize=searchwindowsize,
        expect_type=expect_type,
        mode=mode)

  @decorators.CapabilityLogDecorator(logger)
  def release(self, button, port=0):
    """Release the button for the port specified.

    Args:
        button (str): button to release
        port (int): Which port to release button on, 0 or 1

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button or port values are invalid
    """
    self._check_button_args("release", button, port)

    self._button_list[port].release(button)
    log_message = "released button {}".format(button)
    self.add_log_note(log_message)

  def release_and_expect(self,
                         button,
                         pattern_list,
                         timeout=30.0,
                         searchwindowsize=config.SEARCHWINDOWSIZE,
                         expect_type="log",
                         port=0,
                         mode="any"):
    """Release button, matches pattern_list in loglines as specified by expect_type.

    Args:
        button (str): button to release
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        port (int): which port to send on, 0 or 1
        mode (str): type of expect to run ("any", "all" or "sequential")

    Raises:
        DeviceError: If buttons are not supported on the device or
                     button, port, or expect values are invalid

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an expect. Starts up
        expect queue right before releasing button to catch fast responses.
    """
    return self.do_and_expect(
        self.release, [button], {"port": port},
        pattern_list,
        timeout=timeout,
        searchwindowsize=searchwindowsize,
        expect_type=expect_type,
        mode=mode)

  @decorators.CapabilityLogDecorator(logger, level=None)
  def send(self, command, port=0, slow=False, add_newline=True, newline="\n"):
    """Sends the command to the device on the port (transport) specified.

    Args:
        command (str): to send to the device
        port (int): or transport to send command to
        slow (bool): flag indicating command should be sent byte-by-byte
        add_newline (bool): flag indicating newline should be added to
          command if missing
        newline (str): character to check for and add if missing at the end
          of the command

    Raises:
        DeviceError: if port specified is an invalid value or out of
                     range of the available ports
    """
    self._validate_port(port, self.send.__name__)

    self.add_log_note("wrote command {!r} to port {}".format(command, port))
    command = _ensure_has_newline(command, add_newline, newline)
    if slow or self._force_slow:
      for cmd_byte in command:
        self._send_command_to_device(cmd_byte, port=port)
    else:
      self._send_command_to_device(command, port=port)

  def send_and_expect(self,
                      command,
                      pattern_list,
                      timeout=30.0,
                      searchwindowsize=config.SEARCHWINDOWSIZE,
                      expect_type=line_identifier.LINE_TYPE_ALL,
                      mode=MODE_TYPE_ANY,
                      port=0,
                      slow=False,
                      add_newline=True,
                      newline="\n",
                      command_tries=1,
                      raise_for_timeout=False) -> ExpectResponse:
    r"""Sends the command and expects on the patterns provided.

    Note: this method does not prepend the command with a wakeup character which
    some devices require. The reason this may be needed is because some devices
    go into a sleep state to save energy and will wakeup on receiving the first
    character sent to it which means the character won't get registered into the
    command buffer. This can be dealt with by prepending the command with a nop
    character that won't affect the command being executed in the case that the
    device has already woken up. If there is an issue with this method, try
    adding "\n" in front of the command. E.g. "\nsome_command"

    Args:
        command (str): command to send to the device
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): Seconds to look for the patterns
        searchwindowsize (int): Number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        port (int): Which port to send on, 0 or 1
        slow (bool): flag indicating command should be sent byte-by-byte
        add_newline (bool): flag indicating newline should be added to
          command if missing
        newline (str): character to check for and add if missing at the end
          of the command
        command_tries (int): The number of tries to send the command if it
          times out.
        raise_for_timeout (bool): Raise an exception if the expect times out

    Raises:
        DeviceError: if port specified or other expect arguments are
                     invalid, or timed out and raise_for_timeout was True.

    Returns:
        ExpectResponse: Object with values for the following attributes:
           .index (int): the index of the expected pattern (None if
           timeout).
           .timedout (bool): indicating whether it timed out.
           .time_elapsed (int): number of seconds between start and finish.
           .match (str): re.group of pattern match.
           .before (str): all the characters looked at before the match.
           .after (str):  all the characters after the first matching
           character.
           .remaining (list): remaining patterns not matched
           .match_list (list): re.search pattern MatchObjects

    Note:
        Flushes the expect queue before and after an send.
    """
    for _ in range(command_tries):
      result = self.do_and_expect(
          self.send, [command], {
              "port": port,
              "slow": slow,
              "add_newline": add_newline,
              "newline": newline
          },
          pattern_list,
          timeout=timeout,
          searchwindowsize=searchwindowsize,
          expect_type=expect_type,
          mode=mode)
      if result.timedout:
        continue
      return result

    if result.timedout and raise_for_timeout:
      raise errors.DeviceError(
          "Device {} send_and_expect timed out for command {}".format(
              self._device_name, command))
    return result

  @decorators.CapabilityLogDecorator(logger)
  def set_max_log_size(self, max_log_size):
    """Sets the max_log_size value to the value provided.

    Args:
        max_log_size (int): the max log size to use for log rotation.

    Raises:
        ValueError: if max_log_size is not an integer value
        RuntimeError: if log writer process is not running

    Note:
        A max_log_size of 0 means no log rotation should ever occur.
    """
    if not hasattr(
        self,
        "_log_writer_process") or not self._log_writer_process.is_running():
      raise RuntimeError("Log writer process is not currently running.")
    elif not isinstance(max_log_size, int):
      raise ValueError("Expected 'type<int>' found {}".format(
          type(max_log_size)))

    self._log_writer_process.send_command(log_process.CMD_MAX_LOG_SIZE,
                                          max_log_size)
    while (self._log_writer_process.is_running() and
           not self._log_writer_process.is_command_done()):
      time.sleep(0.001)

  @decorators.CapabilityLogDecorator(logger)
  def start_new_log(self, log_path):
    """Changes log filter and writer to use a new log path provided.

    Args:
        log_path (str): to log file to switch to

    Raises:
        RuntimeError: if LogWriterProcess is not available or running.
    """
    if not hasattr(
        self,
        "_log_writer_process") or not self._log_writer_process.is_running():
      raise RuntimeError("Log writer process is not currently running.")

    if hasattr(self, "_log_filter_process"):
      self._log_filter_process.send_command(log_process.CMD_NEW_LOG_FILE,
                                            log_path)
      while (self._log_filter_process.is_running() and
             not self._log_filter_process.is_command_done()):
        time.sleep(0.001)

    self._log_writer_process.send_command(log_process.CMD_NEW_LOG_FILE,
                                          log_path)
    while (self._log_writer_process.is_running() and
           not self._log_writer_process.is_command_done()):
      time.sleep(0.001)

    # Wait for new log file to appear
    while self._log_writer_process.is_running(
    ) and not os.path.exists(log_path):
      time.sleep(0.1)

  @decorators.CapabilityLogDecorator(logger)
  def transport_jlink_flash(self, image_path, port=0):
    """Calls the J-Link flash method in the transport.

    Args:
      image_path (str): path to the image file to be flashed onto the device.
      port (int): number of the transport to execute the method in.
    """
    self.call(method=jlink_transport.JLinkTransport.flash,
              method_args=(image_path,),
              port=port)

  @decorators.CapabilityLogDecorator(logger)
  def transport_jlink_reset(self, port=0):
    """Calls the J-Link reset method in the transport."""
    self.call(method=jlink_transport.JLinkTransport.reset, port=port)

  @decorators.CapabilityLogDecorator(logger)
  def transport_serial_set_baudrate(self, new_baudrate, port=0):
    """Sets the serial transport baudrate to a different baudrate.

    Args:
      new_baudrate (int): new baudrate to be set, generally 115200 or 921600.
      port (int): number of the transport to set the baudrate for.
    """
    self.call(method=serial_transport.SerialTransport.flush_buffers, port=port)
    self.call(method=serial_transport.SerialTransport.set_property,
              method_args=(transport_properties.BAUDRATE, new_baudrate),
              port=port)
    self.call(method=serial_transport.SerialTransport.flush_buffers, port=port)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def transport_serial_send_xon(self, port=0):
    """Sends the XON control character to the serial transport."""
    self.call(method=serial_transport.SerialTransport.send_xon, port=port)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def transport_serial_send_xoff(self, port=0):
    """Sends the XOFF control character to the serial transport."""
    self.call(method=serial_transport.SerialTransport.send_xoff, port=port)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def transport_serial_send_break_byte(self, port=0):
    """Sends the break control character (Ctrl+C) to the serial transport."""
    self.call(method=serial_transport.SerialTransport.send_break_byte,
              port=port)

  def verify_file_on_transport(self,
                               source_file,
                               destination_path,
                               port=0,
                               method=VERIFY_METHOD_MD5SUM):
    """Verifies source file contents matches destination_path on transport using method.

    Args:
        source_file(path): to compare content to on transport
        destination_path(path): to file to verify on transport
        port(int or str): the transport port to open
        method(str): the method to use to verify destination_path

    Raises:
        DeviceError: If source_file doesn't exist, can't be opened, or
                     the port or method values are invalid or out of range.

    Returns:
        bool: A boolean status indicating verification was successful.

    Note:
        The caller is responsible for preparing the device to receive one
        of the following verification commands::

            md5sum < destination_path >
    """
    self._validate_port(port, self.verify_file_on_transport.__name__)

    if not os.path.exists(source_file):
      raise errors.DeviceError("Device {} verify file on transport failed. "
                               "Source file {} doesn't exist.".format(
                                   self._device_name, source_file))
    elif not isinstance(method, str):
      raise errors.DeviceError(
          "Device {} verify file on transport failed. "
          "Expecting string for method found {} instead.".format(
              self._device_name, type(method)))
    elif method not in _VERIFY_METHODS:
      raise errors.DeviceError("Device {} verify file on transport failed. "
                               "Unknown method value {} expected: {}".format(
                                   self._device_name, method,
                                   ",".join(_VERIFY_METHODS)))

    # Verify md5sum is available on the host computer
    host_md5sum_path = subprocess.check_output(["which", "md5sum"],
                                               stderr=subprocess.PIPE).decode(
                                                   "utf-8", "replace")
    if "md5sum" not in host_md5sum_path:
      raise errors.DeviceError(
          "Device {} echo file to transport failed. "
          "md5sum is not installed on host. "
          "Unable to verify successful file transfer".format(self._device_name))

    try:
      md5sum_results = subprocess.check_output(
          [host_md5sum_path.rstrip(), source_file]).decode("utf-8", "replace")
    except subprocess.CalledProcessError as err:
      raise errors.DeviceError("Device {} echo file to transport failed. "
                               "Unable to retrieve md5sum of {}. "
                               "Error: {!r}".format(self._device_name,
                                                    source_file, err))
    else:
      host_checksum = md5sum_results.split()[0]

    # Get md5sum of destination_path
    checksum_pattern = r"(\w{32})\s+" + destination_path
    missing_pattern = "can't open '" + destination_path + ("': No such file or "
                                                           "directory")
    result = self.send_and_expect(
        "md5sum {}\n".format(destination_path),
        [checksum_pattern, missing_pattern],
        expect_type="response",
        port=port)
    if result.timedout:
      raise errors.DeviceError(
          "Device {} echo file to transport failed. "
          "Unable to verify {} using the md5sum command".format(
              self._device_name, destination_path))
    elif result.index == 0:  # checksum returned for file pattern
      device_checksum = result.match.group(1)
    elif result.index == 1:  # missing file pattern
      device_checksum = "No such file or directory"
    success = host_checksum == device_checksum
    if success:
      log_message = ("verification of {} for port {} was successful".format(
          destination_path, port))
    else:
      log_message = ("verification of {} for port {} failed. "
                     "The host md5sum {!r} != device md5sum {!r}".format(
                         destination_path, port, host_checksum,
                         device_checksum))
    self.add_log_note(log_message)
    return success

  @decorators.CapabilityLogDecorator(logger)
  def xmodem_file_to_transport(self, source_file, port=0):
    """Transfers file to transport specified using the XModem protocol.

    Args:
        source_file(path): to the file to transfer
        port(int or str): the transport port to open

    Raises:
        DeviceError: If source_file doesn't exist, can't be opened, or
                     the port value provided is invalid or out of range.

    Returns:
        bool: A boolean status indicating xmodem transfer was successful.

    Note:
        The caller is responsible for putting the transport into XModem
        transfer mode before calling this method.
    """
    self._validate_port(port, self.xmodem_file_to_transport.__name__)

    if not os.path.exists(source_file):
      raise errors.DeviceError("Device {} xmodem file to transport failed. "
                               "Source file {} doesn't exist.".format(
                                   self._device_name, source_file))

    try:
      stream = io.open(source_file, "rb")
    except IOError as err:
      raise errors.DeviceError("Device {} xmodem file to transport failed. "
                               "Unable to open source file {}. "
                               "Error: {!r}".format(self._device_name,
                                                    source_file, err))
    success = False
    transport = self._transport_processes[port].transport
    modem = xmodem.XMODEM(transport.read, transport.write)
    start_time = time.time()
    try:
      self.close_transport(port=port)
      log_message = "starting xmodem transfer of {} for port {}".format(
          source_file, port)
      self.add_log_note(log_message)
      transport.open()
      success = modem.send(stream)
    finally:
      stream.close()
      log_message = (
          "finished xmodem transfer of {} for port {} in {}s success={}".format(
              source_file, port,
              time.time() - start_time, success))
      self.add_log_note(log_message)
      transport.close()
      self.open_transport(port=port)
    return success

  def add_transport_process(self, transport, **transport_process_kwargs):
    """Add a new transport process to the list of transport processes.

    Args:
        transport(Transport): transport to the device for this process
        **transport_process_kwargs(dict): keyword arguments to the transport
          process
    transport_process_kwargs can be:
        framer(DataFramer): DataFramer derived classes to use to frame
          incoming raw data into raw lines. Defaults to None.
        partial_line_timeout(float): time in seconds to wait before adding
          partial lines to raw_data_queue and log_queue. Defaults to
          transport_process.PARTIAL_LINE_TIMEOUT.
        read_timeout(float): time to wait in seconds for transport reads.
          Defaults to to transport_process._READ_TIMEOUT
        max_read_bytes(int): to attempt to read on each transport read call.
          Defaults to transport_process._MAX_READ_BYTES
        max_write_bytes(int): to attempt to write on each transport write
          call. Defaults to transport_process._MAX_WRITE_BYTES

    Returns:
        int: position of newly added transport process in list of transport
        processes("port")
    """
    self._transport_processes.append(
        transport_process.TransportProcess(
            self._device_name,
            self._mp_manager,
            self._exception_queue,
            self._mp_manager.Queue(),
            self._log_queue,
            transport,
            call_result_queue=self._call_result_queue,
            raw_data_queue=self._raw_data_queue,
            raw_data_id=self._transport_process_id,
            **transport_process_kwargs))
    self._transport_process_id += 1
    return len(
        self._transport_processes) - 1  # The added process is always last

  def _add_transport_processes(self, transport_list, framer_list,
                               partial_line_timeout_list):
    """Create transport processes which handle the given transports.

    Args:
      transport_list(list): list of transports.
      framer_list(list): list of data framers, one per transport.
      partial_line_timeout_list(list): list of float to delay before adding
        a partial line.
    """
    for idx, transport in enumerate(transport_list):
      kwargs = {}
      if idx < len(framer_list):
        kwargs["framer"] = framer_list[idx]
      if idx < len(partial_line_timeout_list):
        kwargs["partial_line_timeout"] = partial_line_timeout_list[idx]
      self.add_transport_process(transport, **kwargs)

  def _add_log_writer_process(self, log_path, max_log_size):
    self._log_writer_process = log_process.LogWriterProcess(
        self._device_name,
        self._mp_manager,
        self._exception_queue,
        self._mp_manager.Queue(),
        self._log_queue,
        log_path,
        max_log_size=max_log_size)

  def _add_log_filter_process(self, parser, log_path):
    if parser is not None:
      self._log_filter_process = log_process.LogFilterProcess(
          self._device_name, self._mp_manager, self._exception_queue,
          self._mp_manager.Queue(), parser, log_path)

  def _check_button_args(self, func_name, button, port, duration=0.0, wait=0.0):
    """Checks that button arguments are valid.

    Args:
        func_name (str): name of the calling function (to include in the
          error message).
        button (str): button name.
        port (int): which port to click on, 0 or 1.
        duration (float): seconds to wait before releasing button.
        wait (float): seconds to wait before returning after pressing the
          button.

    Raises:
        DeviceError: provided button arguments are invalid.
    """
    self._validate_port(port, func_name)

    if not self._button_list:
      raise errors.DeviceError(
          "Device {} {} failed. "
          "Buttons are not supported for this device.".format(
              self._device_name, func_name))
    elif not isinstance(button, str):
      raise errors.DeviceError(
          "Device {} {} failed. "
          "Expecting str for button found {} instead.".format(
              self._device_name, func_name, type(button)))
    elif not isinstance(duration, (int, float)):
      raise errors.DeviceError(
          "Device {} {} failed. "
          "Expecting int or float for duration found {} instead.".format(
              self._device_name, func_name, type(duration)))
    elif not isinstance(wait, (int, float)):
      raise errors.DeviceError(
          "Device {} {} failed. "
          "Expecting int or float for wait found {} instead.".format(
              self._device_name, func_name, type(wait)))
    elif duration < 0.0:
      raise errors.DeviceError(
          "Device {} {} failed. "
          "Invalid duration value {} expected >=0.0".format(
              self._device_name, func_name, duration))
    elif wait < 0.0:
      raise errors.DeviceError("Device {} {} failed. "
                               "Invalid wait value {} expected >=0.0".format(
                                   self._device_name, func_name, wait))
    elif not self._button_list[port].is_valid(button):
      raise errors.DeviceError(
          "Device {} {} failed. "
          "Invalid button {} for port {} expected {}.".format(
              self._device_name, func_name, button, port,
              self._button_list[port].valid_buttons()))

  def _check_expect_args(self, pattern_list, timeout, searchwindowsize,
                         expect_type, mode):
    """Check that expect arguments are valid.

    Args:
        pattern_list (list): list of regex expressions to look for in the
          lines
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")

    Raises:
        DeviceError: provided expect arguments are invalid.
    """
    if not isinstance(pattern_list, list):
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Expecting list for pattern_list ({}) found {} instead.".format(
              self._device_name, repr(pattern_list), type(pattern_list)))
    elif not isinstance(timeout, (int, float)):
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Expecting int or float for timeout found {} instead.".format(
              self._device_name, type(timeout)))
    elif not isinstance(searchwindowsize, int):
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Expecting int for searchwindowsize found {} instead.".format(
              self._device_name, type(searchwindowsize)))
    elif not isinstance(expect_type, str):
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Expecting str for expect_type found {} instead.".format(
              self._device_name, type(expect_type)))
    elif not isinstance(mode, str):
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Expecting str for mode found {} instead.".format(
              self._device_name, type(mode)))
    elif expect_type not in _VALID_EXPECT_TYPES:
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Invalid input for expect_type found {} expecting one of {}.".format(
              self._device_name, expect_type, _VALID_EXPECT_TYPES))
    elif mode not in _VALID_EXPECT_MODES:
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Invalid input for mode found {} expecting one of {}.".format(
              self._device_name, mode, _VALID_EXPECT_MODES))
    elif not pattern_list:
      raise errors.DeviceError(
          "Device {} expect failed. "
          "Expecting at least one pattern in list, found none.".format(
              self._device_name))

    log_message = ("expecting {} patterns from {!r} using "
                   "{} lines and {} search window in {}s".format(
                       mode, pattern_list, expect_type, searchwindowsize,
                       timeout))
    self.add_log_note(log_message)

  def _disable_raw_data_queue(self):
    """Decrement the raw_data_queue user count and disable and flush the raw_data_queue."""
    if self._raw_data_queue_users:
      self._raw_data_queue_users -= 1
    if self._raw_data_queue_users == 0:
      self._toggle_raw_data()
      self._flush_raw_data_queue()

  def _enable_raw_data_queue(self):
    """Increment the raw_data_queue user count and flush and enable the raw_data_queue."""
    self._raw_data_queue_users += 1
    if self._raw_data_queue_users == 1:
      self._flush_raw_data_queue()
      self._toggle_raw_data()

  def _expect(self,
              compiled_list,
              timeout,
              searchwindowsize,
              expect_type,
              mode,
              raise_for_timeout=False):
    """Wait until a message matching the regexps in the list arrives on the data queue.

    Args:
        compiled_list (list): The list of patterns
        timeout (float): seconds to look for the patterns
        searchwindowsize (int): number of the last bytes to look at
        expect_type (str): 'log', 'response', or 'all'
        mode (str): type of expect to run ("any", "all" or "sequential")
        raise_for_timeout (bool): Raise an exception if the expect times out

    Returns:
         ExpectResponse: an expect response

    Raises:
        DeviceError: for a timeout if raise_for_timeout is True.
    """
    actual_searchwindowsize = searchwindowsize * 2
    captured_index = 0
    captured_lines = []
    expected_matches = 1 if mode == "any" else len(compiled_list)
    found_matches = 0
    match_list = []
    window = u""

    start_time = time.time()
    end_time = start_time + timeout
    while time.time() < end_time:
      # Stage 1/4: get next raw line to add to search window
      time_left = end_time - time.time()
      try:
        message = switchboard_process.get_message(
            self._raw_data_queue, timeout=0 if time_left < 0 else time_left)
      except (AttributeError, IOError, queue.Empty, ValueError):
        # manager shutdown or close called
        break

      if message is None:
        continue

      port, line = message
      if not self._identifier.accept(port, line, expect_type):
        continue

      # Stage 2/4: add raw line to search window in searchwindowsize chunks
      captured_lines.append(line)
      line_split_indexes = list(range(0, len(line),
                                      searchwindowsize)) + [len(line)]
      for line_index in range(1, len(line_split_indexes)):
        start_index = line_split_indexes[line_index - 1]
        end_index = line_split_indexes[line_index]
        partial_line = line[start_index:end_index]
        captured_index += len(partial_line)
        window += partial_line
        window = window[-actual_searchwindowsize:]

        # Break if timeout was exceeded
        if time.time() >= end_time:
          break

        # Stage 3/4: find matching pattern from pattern_list in search window
        pattern_list = _get_pattern_list(compiled_list, match_list, mode)
        for pattern in pattern_list:
          match = pattern.search(window)
          if not match:
            continue
          found_matches += 1
          match_list.append(match)

          index = _get_pattern_index(compiled_list, match_list, mode)
          log_message = "found pattern {!r} at index {}".format(
              pattern.pattern, index)
          self.add_log_note(log_message)
          if found_matches != expected_matches:
            # Remove content from window up to match end for next pattern search
            window = window[match.end():]
            break

          # Stage 4/4: expect pattern success, return results
          time_elapsed = time.time() - start_time

          # Compute index into captured_string where match was found
          i = captured_index - len(window) + match.start()

          captured_string = u"".join(captured_lines)
          before = captured_string[:i]
          after = captured_string[i:]

          missing_patterns = _get_missing_patterns(compiled_list, match_list,
                                                   mode)
          remaining_list = _get_pattern_strings(missing_patterns)
          log_message = (
              "mode {} expect completed with {!r} remaining patterns in {}s"
              .format(mode, u", ".join(remaining_list), time_elapsed))
          self.add_log_note(log_message)
          return ExpectResponse(
              index,
              before,
              after,
              match,
              time_elapsed,
              match_list=match_list,
              remaining=remaining_list)

    # Stage 4/4: expect pattern timed out, return results
    time_elapsed = time.time() - start_time

    if match_list:
      match = match_list[-1]
      index = _get_pattern_index(compiled_list, match_list, mode)
    else:
      index = match = None
    before = u"".join(captured_lines)
    after = None

    missing_patterns = _get_missing_patterns(compiled_list, match_list, mode)
    remaining_list = _get_pattern_strings(missing_patterns)
    log_message = (
        "expect timed out after waiting {}s for {!r} remaining patterns".format(
            time_elapsed, u", ".join(remaining_list)))
    self.add_log_note(log_message)

    if raise_for_timeout:
      raise errors.DeviceError(
          "{} expect timed out after waiting {}s for {!r} remaining patterns"
          .format(self._device_name, time_elapsed, ", ".join(remaining_list)))

    return ExpectResponse(
        index,
        before,
        after,
        match,
        time_elapsed,
        timedout=True,
        remaining=remaining_list,
        match_list=match_list)

  def _flush_raw_data_queue(self):
    try:
      while not self._raw_data_queue.empty():
        self._raw_data_queue.get_nowait()
        self._raw_data_queue.task_done()
    except (AttributeError, IOError, queue.Empty):
      # manager shutdown or close called or queue empty
      pass

  def _get_compiled_pattern_list(self, pattern_list):
    """Return compiled regexps objects for the given regex pattern list.

    Args:
        pattern_list (list): list of regular expressions to compile.

    Returns:
        list: compiled regular expressions.

    Raises:
        DeviceError: invalid regular expression provided.
    """
    compiled_list = []
    for pattern in pattern_list:
      try:
        compiled_pattern = re.compile(pattern, re.DOTALL | re.MULTILINE)
        compiled_list.append(compiled_pattern)
      except re.error as err:
        raise errors.DeviceError("Device {} expect failed. "
                                 "Invalid regex pattern {}. Error {!r}".format(
                                     self._device_name, pattern, err))
    return compiled_list

  def _send_command_to_device(self, command, port=0):
    """Send command to port (transport) specified.

    Args:
        command(str): to be sent to port(transport)
        port(int): selecting which transport to send command to
    """
    transport = self._transport_processes[port]
    transport.send_command(transport_process.CMD_TRANSPORT_WRITE, command)

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def delete_last_transport_process(self):
    """Stops and deletes the last transport process in self._transport_processes.

    Note:
        Just stopping a transport process does not delete it.
        All stopped processes are typically reopened after a device reboot.
        The process must be deleted to ensure it is not reopened after a
        device reboot.

    Since we're using process numbers to identify the transport, deleting
    any transport other than the last one will cause some other transports
    in the transport list to shift their transport number by 1, breaking
    their usage. To prevent this, allow deleting only the last process for
    now. The proper solution would be to use some other form of
    identification for processes.

    Raises:
        DeviceError: if there's no transport process to delete.
    """
    if self._transport_processes:
      last_proc_num = len(self._transport_processes) - 1
      self.stop_transport_process(last_proc_num)
      del self._transport_processes[last_proc_num]
    else:
      raise errors.DeviceError("There are no transport processes.")

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def start_transport_process(self, process_num):
    """Start the transport process at position process_num in transport list.

    Args:
        process_num(int): position in self._transport_processes list. This
          position is returned by a prior self.add_transport_process() call.

    Raises:
        DeviceError: if process_num has an invalid value.
    """
    self._validate_port(process_num, self.start_transport_process.__name__)
    if not self._transport_processes[process_num].is_started():
      self._transport_processes[process_num].start()

  @decorators.CapabilityLogDecorator(logger, level=decorators.DEBUG)
  def stop_transport_process(self, process_num):
    """Stop the transport process.

    Args:
        process_num(int): number of transport to stop.
    """
    self._validate_port(process_num, self.stop_transport_process.__name__)
    process = self._transport_processes[process_num]
    if process.is_started():
      process.stop()

  def _start_processes(self):
    for process_num in range(len(self._transport_processes)):
      self.start_transport_process(process_num)
    if hasattr(self, "_log_writer_process"):
      if not self._log_writer_process.is_started():
        self._log_writer_process.start()
    if hasattr(self, "_log_filter_process"):
      if not self._log_filter_process.is_started():
        self._log_filter_process.start()

  def _stop_processes(self):
    """Stop all Switchboard processes."""
    if hasattr(self, "_transport_processes") and self._transport_processes:
      logger.info("{} closing switchboard processes".format(self._device_name))
      while self._transport_processes:
        self.delete_last_transport_process()
    if hasattr(self, "_log_writer_process") and self._log_writer_process:
      if self._log_writer_process.is_started():
        self._log_writer_process.stop()
      delattr(self, "_log_writer_process")
    if hasattr(self, "_log_filter_process") and self._log_filter_process:
      if self._log_filter_process.is_started():
        self._log_filter_process.stop()
      delattr(self, "_log_filter_process")

  def _validate_port(self, port, function_name):
    """Validates the user input (port number).

    Args:
      port(int): port number(position of transport in transport list)
      function_name(str): name of function calling _validate_port

    Raises:
      DeviceError: if port number is not valid OR
    """
    basic_error = "Device {} {} failed. ".format(self._device_name,
                                                 function_name)
    if not isinstance(port, int):
      raise errors.DeviceError(basic_error +
                               "Invalid port type. Expected: int, found: {}."
                               .format(type(port).__name__))
    elif port < 0 or port >= len(self._transport_processes):
      raise errors.DeviceError(
          basic_error +
          "Invalid port number. Expected: [0..{}), found: {}.".format(
              len(self._transport_processes), port))

  def _toggle_raw_data(self):
    for transport in self._transport_processes:
      transport.toggle_raw_data()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TransportBase is the standard GDM device transport interface.

Device transports perform device communications. Device transports are used by
TransportProcess instances.

is_open(): Returns true if transport is open

open(): Opens transport using properties provided during initialization

close(): Closes transport

read(size=1, timeout=None): Reads up to size bytes from device and
returns them or None if no bytes are read and timeout != None.

write(data, timeout=None): Writes bytes to device and returns the number
of bytes written or None if no bytes were written and timeout != None.

get_property(key, value=None): Returns a transport property identified by
key or if the property doesn't exist yet returns value provided.

get_all_properties(): Returns a deepcopy dictionary of all properties keys
and values currently set.

get_property_list(): Returns the list of property keys

set_property(key, value): Sets a transport property identified by key to
the value provided.
"""
import abc
import copy

import six

from gazoo_device.switchboard import transport_properties


class TransportBase(six.with_metaclass(abc.ABCMeta, object)):
  """Interface for transport classes."""

  def __init__(self, auto_reopen=False, open_on_start=True):
    """Initialize the transport interface.

    Args:
        auto_reopen (bool): flag indicating transport should be reopened if
          unexpectedly closed.
        open_on_start (bool): flag indicating transport should be open on
          TransportProcess start.
    """
    if not hasattr(self, "_properties"):
      self._properties = {}
    self._properties.update({
        transport_properties.AUTO_REOPEN: auto_reopen,
        transport_properties.OPEN_ON_START: open_on_start
    })

  @abc.abstractmethod
  def is_open(self):
    """Returns True if the transport is open.

    Returns:
        bool: True if transport is open, False otherwise.
    """

  @abc.abstractmethod
  def _open(self):
    """Open the transport.

    Override in derived classes.

    Note: use properties provided during __init__.
    """

  @abc.abstractmethod
  def _close(self):
    """Close the transport. Override in derived classes."""

  @abc.abstractmethod
  def _read(self, size=1, timeout=None):
    """Returns bytes read up to max_bytes within timeout in seconds specified.

    Args:
        size (int): maximum number of bytes to read within timeout seconds
        timeout (float): maximum seconds to wait to read bytes or
          indefinitely if timeout=None. Default: None

    Returns:
        str: bytes read from transport
             or "" if no bytes were read
             or "" if timeout was reached
    """

  @abc.abstractmethod
  def _write(self, data, timeout=None):
    """Writes the data provided within timeout seconds.

    Args:
        data (str): bytes to be written out within timeout seconds
        timeout (float): maximum seconds to wait to write bytes or
          indefinitely if timeout=None. Default: None

    Returns:
        int: number of bytes written
    """

  def __del__(self):
    """Close the transport on garbage collection."""
    self.close()

  def get_property(self, key, value=None):
    """Returns property matching key specified or value if not set.

    Args:
        key (str): of the property to attempt to retrieve
        value (object): default value to return if the property doesn't exist

    Returns:
        object: The property value for key specified or value if key doesn't
        exist.
    """
    return self._properties.get(key, value)

  def get_all_properties(self):
    """Returns dictionary of property key-value pairs.

    Returns:
        dict: a deepcopy of all properties as a key value pair
    """
    return copy.deepcopy(self._properties)

  def get_property_list(self):
    """Returns a list of property keys that exist and can be set/retrieved.

    Returns:
        list: of property keys that can be set or retrieved.
    """
    return list(self._properties.keys())

  def set_property(self, key, value):
    """Sets the property specified to the value provided.

    Args:
        key (str): of the property to set
        value (object): of the property to set

    Raises:
        KeyError: when key property doesn't exist
    """
    if key not in self._properties:
      raise KeyError("Property {} doesn't exist for {}.".format(
          key, self.__class__.__name__))
    self._properties[key] = value

  def open(self):
    """Open the transport if not open already.

    Note: use properties provided during __init__.
    """
    if self.is_open():
      return None

    self._open()

  def close(self):
    """Closes the transport if it's open."""
    if not self.is_open():
      return None

    self._close()

  def read(self, size=1, timeout=None):
    """Returns bytes read up to max_bytes within timeout in seconds specified.

    Args:
        size (int): maximum number of bytes to read within timeout seconds
        timeout (float): maximum seconds to wait to read bytes or
          indefinitely
                         if timeout=None. Default: None

    Returns:
        str: bytes read from transport
             or None if transport is NOT open
             or "" if no bytes were read
             or "" if timeout was reached

    Note:
        If transport is not opened this method will always return None
    """
    if not self.is_open():
      return None

    return self._read(size, timeout)

  def write(self, data, timeout=None):
    """Writes the data provided within timeout seconds if transport is open.

    Args:
        data (str): bytes to be written out within timeout seconds
        timeout (float): maximum seconds to wait to write bytes or
          indefinitely
                         if timeout=None. Default: None

    Returns:
        int: number of bytes written or None if no bytes were written
    """
    if not self.is_open():
      return None

    return self._write(data, timeout)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common reusable utility functions."""
import multiprocessing
import os
import time
import weakref
from gazoo_device import errors

# Lists of callable functions with no arguments
_run_before_fork_functions = []
_run_after_fork_in_parent_functions = []

# _Sentinel objects used by multiprocessing.util.register_after_fork. This list keeps the objects
# alive by keeping a strong reference to them. multiprocessing.util stores these in a
# weakref.WeakValueDictionary, so if a strong reference is not stored somewhere, the sentinel
# will be destroyed, removing the function registered with it from the list of "after fork"
# functions (b/154659535).
_register_after_fork_sentinels = []


class MethodWeakRef(object):
  """Allows creating weak references to instance methods.

  Note:
      using weakref.ref() on an instance method directly returns a dead
      reference. See https://stackoverflow.com/questions/599430.
  """

  def __init__(self, instance_method):
    """Create a weak reference to an instance (bound) method.

        Args:
            instance_method (method): instance (bound) method.
    """
    self._func = instance_method.__func__
    self._instance_weakref = weakref.ref(instance_method.__self__)

  def __call__(self, *args, **kwargs):
    """Calls the instance method if the instance is still alive."""
    instance = self._instance_weakref()
    if instance is not None:
      self._func(instance, *args, **kwargs)


class _Sentinel(object):
  """Empty object which supports weakrefs."""
  pass


def _default_is_successful(_):
  return True


def not_func(val):
  """Takes an input value and returns bool value after performing a negate operation.

  Args:
      val (object): any python object

  Returns:
      bool: True if bool(val) is False and False if bool(val) is True
  """
  return not bool(val)


def is_true(val):
  """Takes an input value and returns bool value.

  Args:
      val (object): any python object

  Returns:
      bool: bool(val)
  """
  return bool(val)


def retry(func,
          func_args=(),
          func_kwargs=None,
          is_successful=_default_is_successful,
          timeout=10,
          interval=1,
          reraise=True,
          exc_type=errors.CommunicationTimeoutError):
  """Wait until either execution of func() succeeds or timeout is reached.

  Success of execution of func() is determined by is_successful() function,
  which should return True on successful execution of func().

  Args:
      func (function): function to execute
      func_args (tuple): positional arguments to the function
      func_kwargs (dict): keyword arguments to the function
      is_successful (function): function which takes in the result of func()
        and returns whether function execution should be considered
        successful. To indicate success, return True. Defaults to always
        returning True.
      timeout (int): if no run of func() succeeds in this time period, raise
        an error.
      interval (float): how long to wait between retries of func().
      reraise (bool): whether to re-raise exceptions in func() or not. If
        True, will re-raise any exceptions from func(). If False, considers
        execution of func() a failure if an Exception is raised.
        is_successful() will NOT be called if an Exception occurs.
      exc_type (class): type of exception to raise when timeout is reached.
        Note that the class constructor will be called with just 1 argument.

  Returns:
      object: return value of first successful func() call.

  Raises:
      Exception: if timeout is reached. The default exception type is
      errors.TimeoutError. OR if an Exception occurs in func() and reraise is
      False.
  """
  if func_kwargs is None:
    func_kwargs = {}

  tried_times = 0
  start_time = time.time()
  end_time = start_time + timeout

  while time.time() < end_time:
    exception_occurred = False
    tried_times += 1
    try:
      func_result = func(*func_args, **func_kwargs)
    except Exception:
      if reraise:
        raise
      else:
        exception_occurred = True

    if not exception_occurred and is_successful(func_result):
      return func_result

    time.sleep(interval)

  time_elapsed = time.time() - start_time
  raise exc_type("Timeout in {}s. Tried calling {} {} times.".format(
      time_elapsed, func.__name__, tried_times))


def generate_name(_object):
  """Generates a snake_case name (str) for the given object.

  Args:
      _object (object): any python object which has a non-empty __name__
        attribute.

  Raises:
      ValueError: _object does not have __name__ attribute, or __name__ is "".

  Returns:
      str: snake_case name for the object.
  """
  object_name = getattr(_object, "__name__", "")
  if not object_name:
    raise ValueError(
        "Object {} must have a non-empty __name__ attribute.".format(_object))

  if "_" in object_name:  # Presumably already snake_case
    return object_name.lower()
  else:  # Presumably TitleCase
    return title_to_snake_case(object_name)


def get_value_from_json(json_data, key_sequence, raise_if_absent=True):
  """Extracts a value from a JSON dictionary via provided series of keys.

  This function allows easy extraction of a value stored in a nested
  dictionary structure, with validation of the existence of the necessary
  keys at each level of nesting.

  Args:
      json_data (dict): the response object from which the value will be
        extracted.
      key_sequence (list): an ordered list of string keys representing the
        path through the dictionary necessary to reach the desired value.
                           Ex. to extract foo['baz']['a'] the call would look
                             like: get_value_from_json(foo, ['baz', 'a'])  and
                             would return 'some text' given the example data
                             above.
      raise_if_absent (bool): if False, return None if a key is not present.
        if True, raise KeyError if a key is not present.

  Returns:
      object: the requested JSON field.

  Raises:
      KeyError: If the provided response message doesn't contain the necessary
      keys/values AND if raise_if_absent is True.
  """
  current_dict = json_data
  key_text = ""
  for key in key_sequence:
    if key in current_dict:
      current_dict = current_dict[key]
    else:
      if raise_if_absent:
        raise KeyError(
            "Unable to find key '{}' in json_data{}: json_data = {}".format(
                key, key_text, json_data))
      else:
        return None
    key_text += "['{}']".format(key)
  return current_dict


def register_at_fork(before=None, after_in_parent=None, after_in_child=None):
  """Workaround to simulate os.register_at_fork() behavior in all Python versions.

  Args:
      before (func): function to call prior to os.fork() call.
      after_in_parent (func): function to call after os.fork() returns in the
        parent process.
      after_in_child (func): function to call after os.fork() returns in the
        child process.

  Note: For Python >= 3.7, this delegates to os.register_at_fork(). For Python
    < 3.7, the functions need to be run manually via
    common_utils.run_before_fork() prior to os.fork() and
    common_utils.run_after_fork_in_parent() after os.fork() returns in the
    parent process.
      Note that os.fork() is used internally by the "multiprocessing" library:
        multiprocessing.Process().start() and multiprocessing.Manager() both
        use os.fork().
  """
  if hasattr(os, "register_at_fork"):
    os.register_at_fork(
        before=before,
        after_in_parent=after_in_parent,
        after_in_child=after_in_child)
  else:
    global _run_before_fork_functions
    global _run_after_fork_in_parent_functions
    global _register_after_fork_sentinels

    if callable(before):
      _run_before_fork_functions.append(before)
    if callable(after_in_parent):
      _run_after_fork_in_parent_functions.append(after_in_parent)
    if callable(after_in_child):

      def _consume_sentinel_wrapper(_):
        """Consumes the sentinel argument passed into the function."""
        after_in_child()

      sentinel = _Sentinel()
      _register_after_fork_sentinels.append(sentinel)
      # pytype: disable=module-attr
      multiprocessing.util.register_after_fork(sentinel,
                                               _consume_sentinel_wrapper)
      # pytype: enable=module-attr


def run_after_fork_in_parent():
  """Run functions registered to run in parent after os.fork() calls."""
  for func in _run_after_fork_in_parent_functions:
    func()


def run_before_fork():
  """Run functions registered to run before os.fork() calls."""
  for func in _run_before_fork_functions:
    func()


def title_to_snake_case(s):
  """Convert TitleCase string to snake_case.

  Args:
      s (str): TitleCase string.

  Returns:
      str: snake_case string.

  Raises:
      ValueError: provided string contains underscores.

  Note:
      consecutive capital characters are supported ("ABc"), but underscores
      are not.
  """
  if "_" in s:
    raise ValueError(
        "{} is not a TitleCase string (found underscores).".format(s))

  word_starts = [pos for pos in range(len(s)) if _is_new_word(s, pos)
                ] + [len(s)]
  words = [
      s[word_starts[idx]:word_starts[idx + 1]].lower()
      for idx in range(len(word_starts) - 1)
  ]
  return "_".join(words)


def _is_new_word(s, pos):
  """Returns whether a new words starts at s[pos] in a TitleCase string."""
  return (pos == 0 or (s[pos].isupper() and
                       (not s[pos - 1].isupper() or
                        (pos + 1 < len(s) and not s[pos + 1].isupper()))))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for usb information."""
import re
import sys
from gazoo_device.utility import usb_config
from gazoo_device.utility import usb_info_linux
from gazoo_device.utility import usb_info_mac


def find_matching_connections(match_criteria):
  """Returns usb_info instances for connections that match criteria.

  Args:
      match_criteria (dict): dictionary of match criteria.

  Returns:
      list: list of UsbInfo instances that match criteria.

  Raises:
      ValueError: if match criteria contains bad key(s).

  Note:
    Connections first excluded and then included.
    If there is no include regex, all instances that dont match exclude_regex
    are included.

    Dictionary entry: {"<key>": {
                          "include_regex": "<regex>",
                          "exclude_regex": "<regex>""}
                      }
  """
  instances = get_address_to_usb_info_dict().values()
  allowed_keys = usb_config.UsbInfo.get_properties()
  bad_keys = [key for key in match_criteria.keys() if key not in allowed_keys]
  if bad_keys:
    raise ValueError("The keys {} are not allowed. Possible keys: {}".format(
        bad_keys, allowed_keys))
  for attr, entry in match_criteria.items():
    if entry.get("exclude_regex"):
      instances = [
          instance for instance in instances
          if not re.search(entry["exclude_regex"], str(getattr(instance, attr)))
      ]
    if entry.get("include_regex"):
      instances = [
          instance for instance in instances
          if re.search(entry["include_regex"], str(getattr(instance, attr)))
      ]
  return instances


def get_address_to_usb_info_dict():
  """Gets a dictionary of usb devices with all relevent information."""
  if sys.platform == "darwin":
    module = usb_info_mac
  else:
    module = usb_info_linux
  return module.get_address_to_usb_info_dict()


def get_all_serial_connections():
  """Returns a list of all serial connections."""
  usb_info = get_address_to_usb_info_dict()
  return [
      key for key, entry in usb_info.items()
      if entry.product_name not in usb_config.ANDROID_NAMES
  ]


def get_device_info(address):
  """Gets the usb info of a specific device.

  Args:
      address (str): serial path or adb serial.

  Returns:
      object: usb_info instance encoding information for that specific address.
  """
  address_to_info_dict = get_address_to_usb_info_dict()
  if address in address_to_info_dict:
    return address_to_info_dict[address]

  return usb_config.UsbInfo()


def get_vendor_number_from_path(address):
  """Returns the vendor number for the address provided.

  Args:
      address (str): address to find matching vendor id for.

  Raises:
      ValueError: if address not found.

  Returns:
      str: vendor number (e.g. "4012") from address.
  """
  usb_info_inst = get_device_info(address)
  return usb_info_inst.vendor_id


def get_ftdi_interface_from_path(address):
  """Returns the FTDI interface embedded in address provided.

  Args:
      address (str): Address to extract interface from.

  Raises:
      ValueError: if FTDI interface can't be obtained from address.

  Returns:
      str: Interface (e.g. "if02") from address.
  """
  usb_info_inst = get_device_info(address)
  return usb_info_inst.ftdi_interface


def get_product_name_from_path(address):
  """Returns the product name for address provided.

  Args:
      address (str): Address to extract FTDI name from.

  Raises:
      ValueError: if product name can't be obtained from address.

  Returns:
      str: product name from address.
  """
  usb_info_inst = get_device_info(address)
  return usb_info_inst.product_name


def get_other_ftdi_line(address, ftdi_interface):
  """Gets a path to the corresponding ftdi line for that address.

  Args:
     address (str): address for an ftdi based serial.
     ftdi_interface (int): number of the ftdi interface to switch to.

  Returns:
     str: address for other ftdi line.

  Raises:
      ValueError: if a matching ftdi line is not found.
  """
  serial_number = get_serial_number_from_path(address)
  devices = find_matching_connections({
      "serial_number": {
          "include_regex": serial_number
      },
      "ftdi_interface": {
          "include_regex": str(ftdi_interface)
      }
  })
  if devices:
    return devices[0].address
  raise ValueError("No connection found with {} and ftdi interface {}".format(
      serial_number, ftdi_interface))


def get_serial_number_from_path(address):
  """Returns the serial number embedded in address provided.

  Args:
      address (str): Address to extract serial number from.

  Raises:
      ValueError: if FTDI serial number can't be obtained from address.

  Returns:
      str: Serial number from address.
  """
  usb_info_inst = get_device_info(address)
  return usb_info_inst.serial_number


def get_usb_hub_info(device_address):
  """Gets the usb hub information for each communication address of the device.

  Args:
      device_address (str): communication address for device.

  Returns:
      dictionary: keys for usb hub name and port for usb connections.
  """
  usb_info = {}
  usb_info["device_usb_hub_name"] = get_usb_hub_address_from_address(
      device_address)
  usb_info["device_usb_port"] = get_usb_hub_port_from_address(device_address)
  return usb_info


def get_usb_hub_address_from_address(address):
  """Returns the communication address for usb hub the device is connected to.

  Args:
      address (str): communication address of device.

  Returns:
      str: communication address of the usb hub.
      None: if no associated usb hub.
  """
  usb_info_inst = get_device_info(address)
  return usb_info_inst.usb_hub_address


def get_usb_hub_address_from_serial_number(serial_number):
  """Returns the communication address for usb hub the device is connected to.

  Args:
      serial_number (str): device serial_number to associate with port number.

  Returns:
      str: communication address of the usb hub.
      None: if no associated usb hub.
  """
  devices = find_matching_connections(
      {"serial_number": {
          "include_regex": serial_number
      }})
  if devices:
    return devices[0].usb_hub_address
  return None


def get_usb_hub_port_from_address(address):
  """Returns the usb hub port number associated with an address.

  Args:
      address (str): communication Address to associate with port number.

  Returns:
      int: Port number of the usb hub the usb address is attached to.
      None: if no associated port.
  """
  usb_info_inst = get_device_info(address)
  return usb_info_inst.usb_hub_port


def get_usb_hub_port_from_serial_number(serial_number):
  """Returns the usb hub port number associated with an address.

  Args:
      serial_number (str): device serial_number to associate with port number.

  Returns:
      int: Port number of the usb hub the device is attached to.
      None: if no associated port.
  """
  devices = find_matching_connections(
      {"serial_number": {
          "include_regex": serial_number
      }})
  if devices:
    return devices[0].usb_hub_port
  return None



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Used to identify device output lines as Log, Response, or Unknown.

AllUnknownIdentifier (default) which classifies all lines
from all ports as unknown and accepts them for every line_type provided.

AllLogIdentifier classifies all lines from all ports as log lines and only
accepts them for "all" or "log" line_type values.

AllResponseIdentifier classifies all lines from all ports as response lines
and only accepts them for "all" or "response" line_type values.

PortLogIdentifier classifies lines from a particular port as log lines.
All others are responses.

RegexLogIdentifier classifies lines that match a regex as log lines.
All others are unknown.

RegexResponseIdentifier classifies lines that match a regex as response lines.
All others are unknown.
"""
import abc
import re
import six

LINE_TYPE_ALL = "all"
LINE_TYPE_LOG = "log"
LINE_TYPE_RESPONSE = "response"


class LineIdentifier(six.with_metaclass(abc.ABCMeta), object):
  """Base class used to identify device output lines as Log, Response, or Unknown."""

  @abc.abstractmethod
  def accept(self, port, line, line_type):
    """Returns True if line is identified to be part of line_type requested.

    Args:
        port (int): which transport line came from.
        line (str): device output
        line_type (str): line type requested. Options ("all", "log",
          "response")

    Returns:
        bool: whether or not line matches request type.
    """


class AllLogIdentifier(LineIdentifier):
  """Identifies all output as logs."""

  def accept(self, port, line, line_type):
    """Identifies all lines from all ports as log lines."""

    return line_type in [LINE_TYPE_ALL, LINE_TYPE_LOG]


class AllResponseIdentifier(LineIdentifier):
  """Identifies all output as responses."""

  def accept(self, port, line, line_type):
    return line_type in [LINE_TYPE_ALL, LINE_TYPE_RESPONSE]


class AllUnknownIdentifier(LineIdentifier):
  """Identifies all output as unknown."""

  def accept(self, port, line, line_type):
    return True


class PortLogIdentifier(LineIdentifier):
  """Identifies all output from certain ports as logs.

  Note:
      Output from all other ports is considered a reponse.
  """

  def __init__(self, log_ports=None):
    """Identifies lines from log ports as logs.

    Args:
        log_ports (list): ports to identify as logging ports.
    """
    if not log_ports:
      log_ports = [1]
    self.log_ports = log_ports

  def accept(self, port, line, line_type):
    if port in self.log_ports:
      return line_type in [LINE_TYPE_ALL, LINE_TYPE_LOG]
    else:
      return line_type in [LINE_TYPE_ALL, LINE_TYPE_RESPONSE]


class RegexIdentifier(LineIdentifier):
  """Identifies output as match_type if regex matches."""

  def __init__(self, pattern, match_type, use_match=False):
    """Identifies log lines using the log pattern specified."""
    self._pattern = re.compile(pattern)
    self._match_type = match_type
    self._use_match = use_match

  def accept(self, port, line, line_type):
    if self._use_match:
      is_type = self._pattern.match(line)
    else:
      is_type = self._pattern.search(line)

    if line_type == LINE_TYPE_ALL:
      return True
    if line_type == self._match_type and not is_type:
      return False
    elif line_type != self._match_type and is_type:
      return False
    return True


class RegexLogIdentifier(RegexIdentifier):
  """Identifies all output as logs if regex matches else its Unknown."""

  def __init__(self, log_pattern, use_match=False):
    """Identifies all output as logs if regex matches else its Unknown.

    Args:
        log_pattern (str): regex to match to beginning of line.
        use_match (str): must match to beginning of line.
    """
    super(RegexLogIdentifier, self).__init__(
        pattern=log_pattern, match_type=LINE_TYPE_LOG, use_match=use_match)


class RegexResponseIdentifier(RegexIdentifier):
  """Identifies all output as response if regex matches else its Unknown."""

  def __init__(self, response_pattern, use_match=False):
    """Identifies all output as logs if regex matches else its Unknown.

    Args:
        response_pattern (str): regex to match to beginning of line.
        use_match (str): must match to beginning of line.
    """
    super(RegexResponseIdentifier, self).__init__(
        pattern=response_pattern,
        match_type=LINE_TYPE_RESPONSE,
        use_match=use_match)


class MultiportIdentifier(LineIdentifier):
  """Generic line identifier for several ports.

  Each port uses a different identifier.
  MultiportIdentifier delegates to the appropriate identifier.
  """

  def __init__(self, identifiers):
    """Initializes the MultiportIdentifier.

    Args:
        identifiers (list): list of LineIdentifier objects.
    """
    self._identifiers = identifiers

  def accept(self, port, line, line_type):
    """Returns whether the given line should be accepted (True) or discarded (False)."""
    if port >= len(self._identifiers):
      return True
    return self._identifiers[port].accept(
        port=port, line=line, line_type=line_type)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the JLink flashing capability."""
import os
from typing import Dict, List, Optional
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import flash_build_base
import intelhex
import pylink

logger = gdm_logger.get_logger()


class FlashBuildJLink(flash_build_base.FlashBuildBase):
  """J-Link implementation of flash_build capability."""

  def __init__(self,
               device_name: str,
               serial_number: int,
               platform_name: str):
    """Initializes an instance of the FlashBuildJLink capability.

    Args:
      device_name: Device name used for logging.
      serial_number: Device serial number.
      platform_name: The target device's platform name.
    """
    super().__init__(device_name=device_name)
    self._serial_number = serial_number
    self._platform_name = platform_name
    self._jlink = pylink.JLink()

  @decorators.CapabilityLogDecorator(logger)
  def flash_device(self,
                   list_of_files: List[str],
                   expected_version: Optional[str] = None,
                   expected_build_type: Optional[str] = None,
                   verify_flash: bool = True,
                   method: Optional[str] = None) -> None:
    """Flashes the firmware image (.hex file) on the device.

    Args:
      list_of_files: Image files on local host, currently supports flashing
        only one hex file at a time.
      expected_version: Not used.
      expected_build_type: Not used.
      verify_flash: Not used.
      method: Not used.
    """
    del expected_version, expected_build_type, verify_flash, method  # Unused.
    if len(list_of_files) != 1:
      raise ValueError("Only one hex file can be flashed via JLink.")
    image_path = list_of_files[0]
    if not image_path.endswith(".hex"):
      raise ValueError("Only hex type file can be flashed.")
    if not os.path.exists(image_path):
      raise errors.DeviceError(f"Firmware image {image_path} does not exist.")

    self._open_and_halt()

    image = intelhex.IntelHex(os.path.abspath(image_path))
    for segment_start, segment_end in image.segments():
      segment_size = segment_end - segment_start
      segment = image.tobinarray(start=segment_start, size=segment_size)
      self._jlink.flash_write8(segment_start, segment)

    self._reset_and_close()

  def _open_and_halt(self):
    """Opens and connects to the jlink interface and halts the CPU core."""
    self._jlink.open(serial_no=self._serial_number)
    self._jlink.set_tif(pylink.enums.JLinkInterfaces.SWD)
    self._jlink.connect(chip_name=self._platform_name, speed="auto")
    self._jlink.halt()

  def _reset_and_close(self):
    """Resets and closes the jlink interface."""
    self._jlink.reset()
    self._jlink.restart()
    self._jlink.close()

  @decorators.CapabilityLogDecorator(logger)
  def download_build_file(self, remote_build_folder, local_folder):
    """Retrieves the build file(s) from the remote location."""
    raise NotImplementedError(
        "download_build_file is not available in flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def get_defaults(self):
    """Returns a dictionary of default build arguments."""
    raise NotImplementedError(
        "get_defaults is not available in flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def get_firmware_version(self, build_args=None):
    """Returns the firmware version based on the build arguments."""
    raise NotImplementedError(
        "get_firmware_version is not available in flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def get_remote_build_folder(self, build_args=None):
    """Uses the build arguments to determine the remote build folder."""
    raise NotImplementedError(
        "get_remote_build_folder is not available in "
        "flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def extract_build_info(self, build_args=None):
    """Converts the provided build arguments into info about the build."""
    raise NotImplementedError(
        "extract_build_info is not available in flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def latest_verified_build_folder(self):
    """Returns the remote build folder path for the latest verified build."""
    raise NotImplementedError(
        "latest_verified_build_folder is not available in "
        "flash_build_jlink for now.")

  @decorators.CapabilityLogDecorator(logger)
  def upgrade(self,
              build_number: Optional[int] = None,
              build_url: Optional[str] = None,
              build_file: Optional[str] = None,
              forced_upgrade: bool = False,
              latest_verified: bool = False,
              **other_build_args: Dict[str, str]):
    """Upgrade the device based on the provided build arguments.

    Args:
     build_number: Not used.
     build_url: Not used.
     build_file: Local path to the file.
     forced_upgrade: Not used.
     latest_verified: Not used.
     **other_build_args: Not used.
    """
    del build_number, build_url, forced_upgrade, latest_verified  # Unused.
    del other_build_args  # Unused.
    build_args = {"build_file": build_file}
    self.upgrade_over_the_wire(**build_args)

  @decorators.CapabilityLogDecorator(logger)
  def upgrade_over_the_wire(self, **build_args: Dict[str, str]):
    """Using the build arguments, flash the build on the device.

    Args:
      **build_args: Dictionary of build arguments.
    """
    self.flash_device(list_of_files=[build_args["build_file"]])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Raspberry Pi device class."""
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.base_classes import raspbian_device

logger = gdm_logger.get_logger()


class RaspberryPi(raspbian_device.RaspbianDevice):
  """Base Class for RaspberryPi Devices.

  Supports the following functionality:
      --logging
      --shell
      --reboot
  """
  DETECT_MATCH_CRITERIA = {
      detect_criteria.SshQuery.is_rpi: True,
  }
  DEVICE_TYPE = "raspberrypi"
  _OWNER_EMAIL = "gdm-authors@google.com"



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all raspbian devices."""
import re
import time

from gazoo_device import config
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import file_transfer_scp
from gazoo_device.capabilities import shell_ssh
from gazoo_device.switchboard import switchboard
from gazoo_device.utility import deprecation_utils
from gazoo_device.utility import host_utils

logger = gdm_logger.get_logger()

MARKER = "--- GDM Log Marker ---"
LOGGING_FILE_PATH = "/var/log/syslog"
LOG_MARKER_LINE_POS_OR_EMPTY = (
    'grep -n -e "{marker}" {file_path} --text | tail -n 1 | cut -d '
    '":" -f 1').format(
        marker=MARKER, file_path=LOGGING_FILE_PATH)
LOG_MARKER_LINE_POS = (
    "line_num=$({cmd}); [ -z \"($line_num)\" ] && echo 1 || echo"
    " $line_num".format(cmd=LOG_MARKER_LINE_POS_OR_EMPTY))

COMMANDS = {
    "BOOTUP_COMPLETE":
        "sudo systemctl --wait is-system-running",
    "FIRMWARE_VERSION":
        "cat /etc/os-release",
    "INJECT_LOG_MARKER":
        "sudo bash -c 'echo \"{marker}\" >> {file_path}'".format(
            marker=MARKER, file_path=LOGGING_FILE_PATH),
    "KERNEL_VERSION":
        "uname -r",
    "LOGGING":
        "tail -F -n +$({cmd}) {file_path}".format(
            cmd=LOG_MARKER_LINE_POS, file_path=LOGGING_FILE_PATH),
    "MODEL_INFO":
        "cat /proc/device-tree/model",
    "GDM_HELLO":
        "echo 'GDM-HELLO'",
    "REBOOT":
        "sudo reboot",
    "RESET_FAILED":
        "sudo systemctl reset-failed",
    "SERIAL_NUMBER_INFO":
        "cat /proc/cpuinfo",
}

REGEXES = {
    "COMMAND_UNSUPPORTED": r"-bash: \S+: command not found",
    "device_type": "raspberrypi",
    "FIRMWARE_VERSION_REGEX": r"VERSION=\"(\d+ \(\w+\))\"",
    "KERNEL_VERSION_REGEX": r"(.*)",
    "MODEL_INFO_REGEX": r"Raspberry Pi ([^\n]+)",
    "SERIAL_NUMBER_INFO_REGEX": r"Serial\s+: ([^\n]+)"
}

TIMEOUTS = {"GDM_HELLO": 5, "SHELL": 10, "SHUTDOWN": 60, "ONLINE": 120}


class RaspbianDevice(auxiliary_device.AuxiliaryDevice):
  """Base Class for Raspbian Devices."""
  COMMUNICATION_TYPE = "SshComms"
  _COMMUNICATION_KWARGS = {
      "log_cmd": COMMANDS["LOGGING"],
      "key_info": config.KEYS["raspberrypi3_ssh_key"],
      "username": "pi"
  }

  def __init__(self,
               manager,
               device_config,
               log_file_name=None,
               log_directory=None):
    super().__init__(
        manager,
        device_config,
        log_file_name=log_file_name,
        log_directory=log_directory)
    self._commands.update(COMMANDS)
    self._regexes.update(REGEXES)
    self._timeouts.update(TIMEOUTS)

  @decorators.health_check
  def check_device_responsiveness(self):
    """Check if the device is responsive on console.

    Raises:
        DeviceNotResponsiveError: if device is not responsive on console.
    """
    cmd, timeout = self.commands["GDM_HELLO"], self.timeouts["GDM_HELLO"]
    try:
      self.shell(cmd, timeout=timeout)
    except errors.DeviceError as err:
      raise errors.DeviceNotResponsiveError(
          self.name,
          "unable to execute command {!r} on device's shell".format(cmd),
          timeout=timeout,
          details=str(err))

  @decorators.DynamicProperty
  def kernel_version(self):
    """Version of Raspbian kernel.

    Returns:
        str: Raspbian kernel version.
    """
    return self.shell_with_regex(
        self.commands["KERNEL_VERSION"],
        self.regexes["KERNEL_VERSION_REGEX"],
        raise_error=True)

  @decorators.DynamicProperty
  def firmware_version(self):
    """Version of Raspbian.

    Returns:
        str: Raspbian version.
    """
    return self.shell_with_regex(
        self.commands["FIRMWARE_VERSION"],
        self.regexes["FIRMWARE_VERSION_REGEX"],
        raise_error=True)

  @decorators.PersistentProperty
  def health_checks(self):
    """Returns list of methods to execute as health checks."""
    return [
        self.check_device_connected, self.check_create_switchboard,
        self.check_device_responsiveness
    ]

  @decorators.PersistentProperty
  def ip_address(self):
    """Global IP address."""
    return self.communication_address

  @decorators.CapabilityDecorator(shell_ssh.ShellSSH)
  def shell_capability(self):
    return self.lazy_init(
        shell_ssh.ShellSSH,
        self.switchboard.send_and_expect,
        self.name,
        timeout=self.timeouts["SHELL"],
        tries=2)

  @decorators.CapabilityDecorator(file_transfer_scp.FileTransferScp)
  def file_transfer(self):
    """File transfer capability for moving files from and to the device.

    Returns:
        FileTransferScp: file transfer capability using "scp" command.
    """
    return self.lazy_init(
        file_transfer_scp.FileTransferScp,
        ip_address_or_fn=self.ip_address,
        device_name=self.name,
        add_log_note_fn=self.switchboard.add_log_note,
        user=self._COMMUNICATION_KWARGS["username"],
        key_info=self._COMMUNICATION_KWARGS["key_info"])

  @decorators.LogDecorator(logger)
  def recover(self, error):
    """Recovers the device from an error detected by check_device_ready()."""
    if isinstance(error, errors.DeviceNotResponsiveError):
      self.reboot()
    else:
      raise error

  @decorators.LogDecorator(logger)
  def reboot(self, no_wait=False, method="shell"):
    """Reboots the device.

    Verifies device fully boots up afterwards.

    Args:
        no_wait (bool): Return before reboot completes. Default: False
        method (str): reboot technique to use.
    """
    self._inject_log_marker()
    self.switchboard.add_log_note("GDM triggered reboot")
    self.switchboard.send(command=self.commands["REBOOT"])
    if not no_wait:
      self._verify_reboot()

  @decorators.LogDecorator(logger)
  def get_detection_info(self):
    """Gets the persistent and optional attributes of a device during setup.

    Returns:
      tuple: (dict, dict) dictionary of persistent attributes,
              dictionary of optional attributes (set to None).
    """
    self.props = self.props.copy()
    for cmd_name, command in self.commands.items():
      if not cmd_name.endswith("INFO"):
        continue
      regex = self._regexes[cmd_name + "_REGEX"]
      cmd_name = cmd_name.lower()[:-5]  # remove _INFO
      value = self.shell_with_regex(command, regex)
      if re.search(self.regexes["COMMAND_UNSUPPORTED"], value):
        value = "Raspbian detection did not support '{}'".format(command)
      self.props["persistent_identifiers"][cmd_name] = value

    return self.props["persistent_identifiers"], self.props["optional"]

  @classmethod
  def is_connected(cls, device_config):
    """Checks whether or not the device is connected to the computer.

    Args:
        device_config (dict): contains "persistent" dict

    Returns:
        bool: True if device is pingable, False otherwise.

    Notes:
        device_config is typically the device_config from manager.
        It checks if the device is reachable or not.
    """
    ip_address = device_config["persistent"]["console_port_name"]
    return host_utils.is_pingable(ip_address)

  def shell(self,
            command,
            command_name="shell",
            timeout=None,
            port=0,
            include_return_code=False):
    """Sends command and returns response and optionally return code.

    Args:
        command(str): Command to send to the device.
        command_name(str): Identifier for command.
        timeout(float): Time in seconds to wait for device to respond.
        port(int): Which port to send on, 0 or 1.
        include_return_code(bool): flag indicating return code should be
          returned.

    Raises:
        DeviceError: if communication fails.

    Note:
        Can try multiple times as connection can sometimes fail.
        See shell_capability init args for setting the number of retry
        attempts.

    Returns:
        str: If include_return_code is False return the device response to
        the command.
        tuple: If include_return_code is True return the device response and
        return code.
    """
    timeout = timeout or self.timeouts["SHELL"]
    return self.shell_capability.shell(
        command,
        command_name=command_name,
        timeout=timeout,
        port=port,
        include_return_code=include_return_code)

  def shell_with_regex(self,
                       command,
                       regex,
                       regex_group=1,
                       command_name="shell",
                       raise_error=False,
                       tries=1,
                       port=0):
    """Sends a command, searches for a regex in the response, and returns a match group.

    Args:
        command(str): command to issue.
        regex(str): regular expression with one or more capturing groups.
        regex_group(int): number of regex group to return.
        command_name(str): command name to appear in log messages.
        raise_error(bool): whether or not to raise error if unable to find a
          match.
        tries(int): how many times to try executing the command before
          failing.
        port(int): which port to send the shell command to.

    Returns:
        str: value of the capturing group with index 'regex_group' in the
        match.

    Raises:
        DeviceError: if command execution fails OR
                     couldn't find the requested group in any of the
                     responses.
    """
    return self.command_with_regex(
        command,
        regex,
        self.shell,
        regex_group=regex_group,
        raise_error=raise_error,
        tries=tries,
        command_name=command_name,
        port=port)

  @decorators.CapabilityDecorator(switchboard.SwitchboardDefault)
  def switchboard(self):
    """Instance for communicating with the device."""
    switchboard_name = self._get_private_capability_name(
        switchboard.SwitchboardDefault)
    if not hasattr(self, switchboard_name):
      switchboard_kwargs = self._COMMUNICATION_KWARGS.copy()
      switchboard_kwargs.update({
          "communication_address": self.communication_address,
          "communication_type": self.COMMUNICATION_TYPE,
          "log_path": self.log_file_name,
          "device_name": self.name,
          "event_parser": None})
      setattr(self, switchboard_name,
              self.manager_weakref().create_switchboard(**switchboard_kwargs))

    return getattr(self, switchboard_name)

  def _ensure_device_goes_offline(self, timeout=None):
    """Ensure device is no longer pingable over ssh.

    Args:
        timeout(float): Time in seconds to wait for device to respond.

    Raises:
        DeviceError: Deviced failed to go offline before the timeout
    """
    timeout = timeout or self.timeouts["SHUTDOWN"]
    start_time = time.time()
    max_disconnect_time = start_time + timeout
    count = 0
    while time.time() < max_disconnect_time:
      if not host_utils.is_pingable(self.ip_address):
        count += 1  # Ensure device is really offline not just a blip
      else:
        count = 0
      if count == 2:
        logger.info("{} offline in {}s.".format(self.name,
                                                int(time.time() - start_time)))
        # close ssh transport as the ssh connection is disconnected.
        self.switchboard.close_all_transports()
        time.sleep(5)  # to ensure offline
        return
      time.sleep(.5)

    raise errors.DeviceError("Failed to go offline within {}s.".format(timeout))

  def _inject_log_marker(self):
    """Add GDM log marker to / var / log / syslog.

    This is to prevent reading stale logs.

    Note:
        Device logs are read starting with the last log marker(if present).
    """
    self.shell(self.commands["INJECT_LOG_MARKER"])

  def _ensure_device_is_online(self, timeout=None):
    """Ensure device is online and configs are fully loaded.

    Args:
        timeout(float): Time in seconds to wait for device to respond.

    Raises:
        DeviceError: Device failed to come online before the timeout.
    """
    timeout = timeout or self.timeouts["ONLINE"]
    start_time = time.time()
    max_disconnect_time = start_time + timeout
    while time.time() < max_disconnect_time:
      if host_utils.is_pingable(self.ip_address):
        # There's a delay between the device being responsive to ping
        # and being able to open SSH connections
        time.sleep(10)
        self.switchboard.open_all_transports()
        break
      time.sleep(.5)
    else:
      raise errors.DeviceError("{} failed to become pingable in {}s.".format(
          self.name, timeout))

    output = "Device still offline"
    while time.time() < max_disconnect_time:
      # Ensure the BOOTUP_COMPLETE command is only sent once
      boot_up_complete_timeout = max_disconnect_time - time.time()
      try:
        output, return_code = self.shell(
            self.commands["BOOTUP_COMPLETE"],
            timeout=boot_up_complete_timeout,
            include_return_code=True)
        if return_code == 0:  # command executed
          logger.info("{} online in {}s".format(self.name,
                                                int(time.time() - start_time)))
          return
        self.shell(self.commands["RESET_FAILED"])
      except errors.DeviceError:
        logger.debug(
            "{} failed to respond to {!r}.".format(
                self.name, self.commands["BOOTUP_COMPLETE"]),
            exc_info=True)
      time.sleep(.5)
    raise errors.DeviceError(
        "Failed to come online and respond to {!r} in {}s. Response: {}".format(
            self.commands["BOOTUP_COMPLETE"], timeout, output))

  def _verify_reboot(self):
    """Verifies reboot actually occurred."""
    self._ensure_device_goes_offline()
    self._ensure_device_is_online()

  def _list_properties_dynamic_raspbian(self):
    dyn_list = ["firmware_version", "kernel_version"]
    return set(dyn_list)

  def __del__(self):
    self._log_object_lifecycle_event("__del__")
    if hasattr(self, "close"):
      self.close()


deprecation_utils.add_deprecated_attributes(
    RaspbianDevice,
    [("send_file_to_device", "file_transfer.send_file_to_device", True),
     ("recv_file_from_device", "file_transfer.recv_file_from_device", True),
     ("do_and_expect", "switchboard.do_and_expect", True),
     ("expect", "switchboard.expect", True),
     ("send", "switchboard.send", True),
     ("send_and_expect", "switchboard.send_and_expect", True)])
